try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from sys import version_info
from socket import gethostname
from psutil import virtual_memory
from omfit_classes.utils_fusion import is_device

__all__ = ['OMFITtoksearch', 'TKS_MdsSignal', 'TKS_PtDataSignal', 'TKS_Aligner', 'TKS_OMFITLoader']

# https://vali.gat.com/toksearch/toksearch

# CONSTANTS
LOGIN_NODE_LIMIT = 5  # max gb fetch on iris login node, set by policy
SERVER_MEM = {'iris': 120, 'saga': 120}  # TO DO: need to use remote_sysinfo to find this
SERVER_NAME = gethostname().split('.')[0]
PYTHON_VERSION = version_info[0]
TOKSEARCH_VERSION = 'toksearch/release'

allowed_partitions = set(['short', 'medium', 'long', 'nodes'])

bash_script = """#!/bin/sh -l
#SBATCH --job-name=toksearch_omfit_remote
#SBATCH --output=toksearch_omfit_remote.out
#SBATCH --partition={partition}
#SBATCH --ntasks={n}
#SBATCH --nodes=1
#SBATCH --mem={mem}G
#SBATCH --time={hours}:{minutes}:00
CUDA_VISIBLE_DEVICES=''
module purge
module load %s
python3 toksearch_python.py
"""

executables = {'iris': 'sbatch %s',
               'saga': 'sbatch %s',
               'iris_batch': bash_script % (TOKSEARCH_VERSION),
               'saga_batch': bash_script % (TOKSEARCH_VERSION)}

login_nodes = set(('irisa', 'irisb', 'irisc', 'irisd'))

toksearchQueryString = """
def toksearchQueryRun(shots, signals, server_info, datasets=None, aligners=None, functions=None, where=None, keep=None,
                      compute_type='ray', return_data='by_shot', warn=True, use_dask=True, **compute_kwargs):
    from sys import path
    path.insert(0, '/fusion/projects/omfit-results/toksearch')
    from toksearchQuery import toksearchQuery
    return toksearchQuery(shots, signals, server_info, datasets, aligners, functions, where, keep,
                          compute_type, return_data, warn, use_dask, **compute_kwargs)
"""


# helper functions
def format_time_limit(hours=1,minutes=0):
    #avoid invalid times
    minutes = int(minutes)
    hours = int(hours)
    if minutes < 0 or minutes > 60:
        raise ValueError('Minutes for time limit must be an integer between [0,60), you entered: %s'%(minutes))
    if hours < 0 or hours > 24:
        raise ValueError('Hours for time limit must be an integer between [0,24], you entered: %s'%(hours))
    hours = '%02d'%(hours)
    minutes = '%02d'%(minutes)
    return hours, minutes

def server_info():
    from resource import getrusage, RUSAGE_SELF
    return {'start_memory_usage': getrusage(RUSAGE_SELF).ru_maxrss,
            'omfit_server': SERVER_NAME,
            'memory_available': virtual_memory().available,
            'python_version': PYTHON_VERSION}


def toksearch_wrapper(basename, *args, **kw):
    kw = {} if kw is None else kw
    args = [] if args is None else args
    return [basename, args, kw]


def TKS_MdsSignal(*args, **kw):
    return toksearch_wrapper('MdsSignal', *args, **kw)


def TKS_PtDataSignal(*args, **kw):
    return toksearch_wrapper('PtDataSignal', *args, **kw)


def TKS_OMFITLoader(*args, **kw):
    return toksearch_wrapper('OMFITLoader', *args, **kw)


def TKS_Aligner(align_with, **kw):
    '''
    Function that takes in a signal name, and keyword arguments and puts them in correct format that the toksearch query method expects.
    The signal specified is the one that the dataset is intended to be aligned with.

    :param align_with: A string respresenting the name of the signal name in 'signals' that the dataset is to be aligned with respect to.
    '''
    return (align_with, kw)


def WARN(size):
    message = "WARNING: It is estimated that you are requesting approx. %.2f GB of data. Are you sure you want to do this?" % (size / (1024 ** 3))
    if OMFITaux.get('GUI', None):
        from omfit_classes.OMFITx import Dialog
        return Dialog(message, ['Yes', 'No'], 'warning', "TOKSEARCH REQUESTING LARGE FILE SIZE")
    else:
        printw(message)


def NOTIFY(size):
    if SERVER_NAME in login_nodes:
        message = "ERROR: You are trying to pull back approx. %.2f GB of data. This will put you beyond the allowed limit of %s GB of accumulated data on a login node. \nPlease reduce your query size, or use SLURM que or another machine to run OMFIT." % (size * 1e-9, LOGIN_NODE_LIMIT)
    else:
        message = "ERROR: You are trying to pull back approx. %.2f GB of data. Which is beyond the available limit on your machine." % (size / (1024 ** 3))
    from omfit_classes.OMFITx import Dialog
    return Dialog(message, ['Ok'], 'error', "TOKSEARCH MEMORY LIMIT HIT")


def predictTokUse(mem_required):
    return round(mem_required * 4 + 25)


class OMFITtksValue():
    '''
    Wrapper for toksearch class, to be interfaced as an OMFIT MDS plus class.
    '''

    def __init__(self, record, server=None, treename=None, shot=None, TDI=None, quiet=False):
        self.record = record
        self.server = server
        self.treename = treename
        self.shot = shot
        self.TDI = TDI
        if shot is None or TDI is None:
            self.is_dataset = False
        else:
            self.is_dataset = isinstance(self.record[self.shot][self.TDI], xarray.core.dataset.Dataset)

    def data(self, sig_name=None):
        try:
            if self.is_dataset:
                if sig_name is None:
                    return self.record[self.shot][self.TDI].variables
                else:
                    return self.record[self.shot][self.TDI][sig_name]
            else:
                return self.record[self.shot][self.TDI]['data']
        except Exception as ex:
            print("Error while fetching data: " + str(ex))

    def dim_of(self, dim):
        if self.is_dataset:
            dim_name = list(self.record[self.shot][self.TDI].dims.keys())[dim]
        else:
            dim_name = self.record[self.shot][self.TDI]['dims'][dim]
        return self.record[self.shot][self.TDI][dim_name]

    def units(self, sig_name=None):
        try:
            if self.is_dataset:
                if sig_name is None:
                    return self.record[self.shot][self.TDI].attrs['units']
                else:
                    return self.record[self.shot][self.TDI][sig_name].attrs['units']
            else:
                return self.record[self.shot][self.TDI]['units']['data']
        except Exception as ex:
            return '?'

    def units_dim_of(self, dim):
        try:
            if self.is_dataset:
                dim_name = list(self.record[self.shot][self.TDI].dims.keys())[dim]
                return self.record[self.shot][self.TDI][dim_name].attrs['units']
            else:
                dim_name = self.record[self.shot][self.TDI]['dims'][dim]
                return self.record[self.shot][self.TDI]['units'][dim_name]
        except Exception as ex:
            return '?'

    def xarray(self):
        """
        :return: DataArray with information from this node
        """

        data = self.data()
        dims = ['dim_%d' % k for k in range(data.ndim)]
        clist=[]
        for k, c in enumerate(dims):
            clist.append(self.dim_of(k))

        if data.shape != tuple([len(k) if np.ndim(k) == 1 else k.shape[ik] for ik,k in enumerate(clist)]):
            dims=dims[::-1]
            clist=clist[::-1]

        coords = {}
        for k, c in enumerate(dims):
            if np.ndim(clist[k]) == 1:
                ck=c
                coords[ck]=([c],clist[k],{'units': self.units_dim_of(k)})
            else:
                ck=c+'_val'
                coords[ck]=(dims, clist[k],{'units': self.units_dim_of(k)})

        xdata = DataArray(data, dims=dims, coords=coords, attrs={'units': self.units()})
        return xdata


def toksearchQueryRemote(serverPicker, shots, signals, datasets=None, aligners=None, functions=None, where=None, keep=None,
                         compute_type='ray', mem_requested=30, return_data='by_shot', use_dask=True,
                         load_data=True, queue_args={}, warn=True, **compute_kwargs):
    '''

    This function creates a toksearch query on a designated server (server must support toksearch). Takes in a list shot
    numbers and signal point names.

    :param serverPicker: (string) A string designating the server to create the toksearch query on.

    :param shots: A list of shot numbers (ints) to be fetched

    :param signals: A dict where each key corresponds to the signal name returned by toksearch, and each entry is a list
        which corresponds to a signal object fetched by toksearch. The first element of the list is the string corresponding to each signal name, i.e. 'PtDataSignal', the 2nd and 3rd entries are the args (list), and keyword args (dict) respectively. Ex) ['PtDataSignal',['ip'],{}] corresponds to a fetch for
        PtData 'ip' signal.

    :param datasets: A dict representing xarray datasets to be created from fetched signals.

    :param aligners: A dict where the keys are name of the dataset to align and the entries are a corresponding list of Aligners
    :param functions: A list of functions or executable strings to be executed in the toksearch mapping stage

    :param where: (string) An evaluatable string (executed in namespace of record) which should return a boolean when
        the record should be returned by toksearch query. This shall be used when trying to filter out shots by certain
        criteria. i.e. return False when you wish to filter out a shot, when string evaluates to True.

    :param keep: A list of strings pertaining to which attributes (signal,dataset etc.) of each record to be
        returned by toksearch query default: returns all attrs in namespace record

    :param compute_type : (string) Type of method to be used to run the pipeline. Options: 'serial','spark','ray'
                                  compute_type='ray' gives better memory usage and parallelization
    :param return_data: (string) A string pertaining on how the data fetched by toksearch should be structured.
        Options: 'by_shot','by_signal'. 'by_shot' will return a dictionary with shots as keys, with each record namespace
        stored in each entry. 'by_signal' will return a dictionary organized with the union of all record attrs as keys,
        and a dictionary organized by shot numbers under it. default: 'by_shot'
        NOTE: When fetching 'by_signal' Datasets will concatinated together over all valid shots.

    :param use_dask: (bool) Boolean of whether to load datasets using dask. Loading with dasks reduces the amount of
        RAM used by saving the data to disk and only loading into memory by chunks. default: False

    :param load_data: return data or return list of files for the user to handle

    :param **compute_kwargs: keyword arguments to be passed into the toksearch compute functions

    :return: data or return list of files for the user to handle depending on `load_data` switch
    '''
    queued_system = {'clean_after': True}

    queue_args.setdefault('partition', 'short' if is_server(serverPicker, 'iris') else 'nodes')  # type of queue resource
    queue_args.setdefault('n', 10)  # number of cpus
    queue_args.setdefault('mem', int(mem_requested))  # amount of RAM allocated
    queue_args.setdefault('hours',1)
    queue_args.setdefault('minutes',0)
    queue_args['hours'], queue_args['minutes'] = format_time_limit(queue_args['hours'],queue_args['minutes'])

    if not queue_args['partition'] in allowed_partitions:
        raise RuntimeError("ERROR: Partition type %s not supported. Choose: 'short','medium' or 'long'" % (queue_args['partition']))

    if is_server(serverPicker, 'saga'):
        queued_system['script'] = (executables['saga_batch'].format(**queue_args), 'batch.sh')
        queued_system['queued'] = True
        queued_system['std_out'] = 'toksearch_omfit_remote.out'
    else:
        raise NotImplementedError(str(serverPicker) + " is not a supported server.")
        printe("Error: " + str(serverPicker) + " is not a supported server.")
        return

    namespace = {'shots': shots,
                 'signals': signals,
                 'functions': functions,
                 'where': where,
                 'keep': keep,
                 'aligners': aligners,
                 'datasets': datasets,
                 'compute_type': compute_type,
                 'use_dask': use_dask,
                 'return_data': return_data.lower(),
                 'warn': warn,
                 'server_info': server_info()}
    namespace.update(compute_kwargs)

    data_files = ['toksearch.toks']
    if return_data == 'by_signal':
        data_files.extend(ds + '.tokds' for ds in namespace['datasets'])

    from omfit_classes.OMFITx import remote_python
    from time import time
    date_time = now("%Y-%m-%d_%H_%M_%S_") + (str(hash(time()))[:6])
    run_folder = 'toksearch_run_' + date_time + '.tokrun'
    workDir = SERVER['localhost']['workDir']
    data_files, mem_usage, data_size, ret_code = remote_python(None,
                                                               python_script=(toksearchQueryString, 'toksearch_python.py'),
                                                               target_function='toksearchQueryRun',
                                                               namespace=namespace,
                                                               executable=executables[serverPicker],
                                                               server=SERVER[serverPicker]['server'],
                                                               workdir=workDir + run_folder,
                                                               remotedir=SERVER[serverPicker]['workDir'] + run_folder,
                                                               tunnel=SERVER[serverPicker]['tunnel'],
                                                               outputs=data_files,
                                                               **queued_system)
    move_data = OMFITsessionDir not in workDir
    data_path = OMFITsessionDir + '/' + run_folder
    if move_data:
        if not os.path.exists(data_path):
            os.makedirs(data_path)

    if load_data:
        with open('toksearch.toks', 'rb') as f:
            data = pickle.load(f)
        if return_data == 'by_signal' and use_dask:
            if move_data:
                shutil.move('toksearch.toks', data_path + '/toksearch_' + date_time + '.toks')
            for ds in namespace['datasets']:
                try:
                    if move_data:
                        shutil.move(ds + '.tokds', data_path + '/' + ds + '.tokds')
                    data[ds] = xarray.open_dataset(data_path + '/' + ds + '.tokds', chunks={'shot': 10})
                except Exception as ex:
                    print("FILE %s MISSING" % (ds + '.tokds'))
                    print(ex)
    else:
        if move_data:
            moved_files = []
            for file in data_files:
                shutil.move(file, data_path + '/toksearch_' + date_time + '.toks')
                moved_files.append(data_path + '/toksearch_' + date_time + '.toks')
            data = moved_files
        else:
            data = data_files

    if move_data:
        os.chdir(data_path)
        shutil.rmtree(workDir + run_folder)

    return data, mem_usage, data_size, ret_code


class OMFITtoksearch(SortedDict):
    '''
        This class is used to query from database through tokesearch API

        :param serverPicker: (string)A string designating the server to create the toksearch query on.

        :param shots: A list of shot numbers (ints) to be fetched

        :param signals: A dict where each key corresponds to the signal name returned by toksearch, and each entry is a list
            which corresponds to a signal object fetched by toksearch. The first element of the list is the string
            corresponding to each signal name, i.e. 'PtDataSignal', the 2nd and 3rd entries are the args (list), and keyword
            args (dict) respectively. Ex) ['PtDataSignal',['ip'],{}] corresponds to a fetch for
            PtData 'ip' signal.

        :param datasets: A dict representing xarray datasets to be created from fetched signals.

        :param aligners: A dict where the keys are name of the dataset to align and the entries are a corresponding list of Aligners

        :param functions: A list of functions or executable strings to be executed in the toksearch mapping stage

        :param where: (string) An evaluatable string (executed in namespace of record) which should return a boolean when
            the record should be returned by toksearch query. This shall be used when trying to filter out shots by certain
            criteria. i.e. return False when you wish to filter out a shot, when string evaluates to True.

        :param keep: A list of strings pertaining to which attributes (signal,dataset etc.) of each record to be
            returned by toksearch query default: returns all attrs in namespace record

        :param compute_type: (string) Type of method to be used to run the pipeline. Options: 'serial','spark','ray'
                                      compute_type='ray' gives better memory usage and parallelization

        :param return_data: (string) A string pertaining on how the data fetched by toksearch should be structured.
            Options: 'by_shot','by_signal'. 'by_shot' will return a dictionary with shots as keys, with each record namespace
            stored in each entry. 'by_signal' will return a dictionary organized with the union of all record attrs as keys,
            and a dictionary organized by shot numbers under it. default: 'by_shot'
            NOTE: When fetching 'by_signal' Datasets will concatinated together over all valid shots.

        :param warn: (bool) If flag is true, the user will be warned if they are about to pull back more than 50% of their available memory and can respond accordingly. This is a safety precaution when pulling back large datasets that may cause you to run out of memory. (default: True).

        :param use_dask: (bool) If flag is True then created datasets will be loaded using dask. Loading with dasks reduces the amount of
            RAM used by saving the data to disk and only loading into memory by chunks. (default: False)

        :param load_data: (bool) If this flag is False, then data will be transferred to disk under OMFIT current working directory, but the data will not be loaded into memory (RAM) and thus the OMFITtree will not be updated. This is to be used when fetching data too large to fit into memory. (default True).

        :param **compute_kwargs: keyword arguments to be passed into the toksearch compute functions
    '''

    def __save_kw__(self):
        return self.args

    def __init__(self, shots, signals, server='saga', datasets=None, aligners=None, functions=None, where=None, keep=None, compute_type='spark', return_data='by_shot', warn=True, use_dask=False, load_data=True, queue_args={}, **compute_kwargs):
        SortedDict.__init__(self)
        self.kw = compute_kwargs
        self.kw['serverPicker'] = server
        self.kw['shots'] = shots
        self.kw['signals'] = signals
        self.kw['datasets'] = datasets
        self.kw['aligners'] = aligners
        self.kw['compute_type'] = compute_type
        self.kw['functions'] = functions
        self.kw['where'] = where
        self.kw['keep'] = keep
        self.kw['use_dask'] = use_dask
        self.kw['return_data'] = return_data.lower()
        self.kw['load_data'] = load_data
        self.kw['queue_args'] = queue_args.copy()
        self.kw['warn'] = warn
        self.args = self.kw.copy()
        self.kw['mem_requested'] = 60
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        kw = self.kw.copy()
        objs_as_dict, memory_usage, serial_mem, ret_code = toksearchQueryRemote(**kw)
        self.dynaLoad = False
        if not kw['load_data']:
            return objs_as_dict
        if ret_code == 1:  # pulling too much data
            ans = NOTIFY(serial_mem)
            self.dynaLoad = True
        elif ret_code == 3 and kw['warn']:  # will warn if self.warn == True
            pull_data = WARN(serial_mem)
            if pull_data == 'Yes':
                mem_needed = predictTokUse(serial_mem)
                kw['mem_requested'] = min(mem_needed, SERVER_MEM[kw['serverPicker']])
                kw['warn'] = False
                objs_as_dict, memory_usage, serial_mem, ret_code = toksearchQueryRemote(**kw)
                if ret_code == 1:
                    ans = NOTIFY(serial_mem)
                    self.dynaLoad = True
                else:
                    self.update(objs_as_dict)
            else:
                self.dynaLoad = True
        else:
            self.update(objs_as_dict)

    @dynaLoad
    def __call__(self, *args, **kw):
        if args or kw:
            return OMFITtksValue(self, *args, **kw)

############################################
if '__main__' == __name__:
    test_classes_main_header()
    if SERVER['saga']['workDir'] is None:
        printe(_using_toksearch_outside_framework_warning)
    filter_func = "not record['errors']"
    OMFIT['ts']=OMFITtoksearch([133221], {'ip':TKS_PtDataSignal('ip')}, where=filter_func)
    #successful creation of OMFIT object (no connections tested)



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

__all__ = ['OMFITerror', 'OMFITobjectError', 'OMFITexpressionError']


class OMFITerror(object):
    def __init__(self, error='Error', traceback=None):
        self.error = error
        self.traceback = traceback

    def __repr__(self):
        return self.error


class OMFITobjectError(OMFITerror, OMFITobject, SortedDict):
    """
    This class is a subclass of OMFITobject and is used in OMFIT
    when loading of an OMFITobject subclass object goes wrong
    during the loading of a project.

    Note that the orifinal file from which the loading failed is not lost
    but can be accessed from the .filename attribute of this object.
    """

    def __init__(self, filename, className=None, traceback=None, **kw):
        SortedDict.__init__(self)
        OMFITobject.__init__(self, filename, **kw)
        errorMessage = 'Error with OMFITobject ' + os.path.split(filename)[1]
        if className is not None:
            errorMessage += ' of class ' + className
        self.error = kw.pop('error', errorMessage)
        self.className = className
        self.traceback = traceback

    def __tree_repr__(self):
        return os.path.split(self.filename)[-1] + '  ' + self.error, []


class OMFITexpressionError(OMFITerror):
    def __init__(self, error='Expression Error', **kw):
        OMFITerror.__init__(self, error, **kw)



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.utils_fusion import is_device, tokamak

import numpy as np

try:
    import pyodbc
except Exception as excp:
    warnings.warn('No RDB support: ' + repr(excp))

__all__ = ['OMFITrdb', 'available_EFITs', 'translate_RDBserver', 'set_rdb_password', 'get_rdb_password']

_RDBserverDict = {
    # dictionary to translate mnemonic names to RDB servers stored in MainSettings['SERVER']
    'd3drdb': 'd3drdb',
    'atlas.gat.com': 'd3drdb',
    'DIIID': 'd3drdb',
    'DIII-D': 'd3drdb',
    'D3D': 'd3drdb',
    'gat': 'd3drdb',
    'd3dpub': 'huez',
    'huez.gat.com': 'huez',
    'CMOD': 'alcdb2',
    'C-Mod': 'alcdb2',
    'EAST': 'east_database',
    'east': 'east_database',
}


def printq(*args):
    """Shortcut so that the topic for the file can be set consistently and easily"""
    printd(*args, topic='omfit_rdb')


def translate_RDBserver(server, servers=_RDBserverDict):
    """This function maps mnemonic names to real RDB servers"""
    server = tokamak(server)
    if server.split(':')[0] not in list(OMFIT['MainSettings']['SERVER'].keys()):
        score_table = []
        for item in list(servers.keys()):
            m = difflib.SequenceMatcher(None, item.lower(), server.lower())
            score_table.append([m.ratio(), item, servers[item]])

        # return best match
        scores = list(map(eval, np.array(score_table)[:, 0]))
        if max(scores) > 0.80:
            server = score_table[np.array(scores).argmax()][2]
        else:
            raise OMFITexception(
                "Server '"
                + server
                + "' was not recognized.\nYou need to add '"
                + server
                + "' to the list of servers in OMFIT['MainSettings']['SERVER']"
            )

    return server


def _get_rdb_credential(server):
    """
    Provides the name of the encrypted credentials file
    :param server: string
        SQL database server. Each server can have its own file.
    :return: string
        Credential name
    """
    return 'rdb_{}:0'.format(translate_RDBserver(server))


def _read_rdb_password_from_file(server=None, filename=None):
    """
    Attempts to read a typical login file and get the username and password
    :param server: string
        Sets the default filenames to search
    :param filename: string
        Override the filename to search
    :return: (string, string)
        Username and password if successful or ('', '') if unsuccessful
    """
    username = ''
    password = ''
    files = {
        'huez': ['.pgdb', '.pgpass'],
        'd3drdb': ['D3DRDB.sybase_login', 'd3drdb.sybase_login'],
        'd3dsqlsrvr': ['D3DRDB.sybase_login', 'd3drdb.sybase_login'],
        'd3sqlsrvr': ['D3SQLSRVR.sybase_login', 'd3sqlsrvr.sybase_login'],
        'east_database': ['east_database.sybase_login'],
    }.get(translate_RDBserver(server), [''])
    files = [filename] if filename is not None else files

    def parse_login1(ff):
        return map(lambda x: x.strip(), ff.read().split(':')[-2:])

    def parse_login2(ff):
        return map(lambda x: x.strip(), ff.readlines())

    parsers = [parse_login1, parse_login2]
    printq('Searching for {} login info in files {}'.format(server, files))
    for login_file in files:
        login_file = os.environ['HOME'] + os.sep + login_file
        if os.path.exists(login_file):
            for i, parser in enumerate(parsers):
                try:
                    printq('  Attempt {}: read {} login from {}'.format(i, server, login_file))
                    with open(login_file, 'r') as _f:
                        username, password = parser(_f)
                    printq('  Success: username {} from file on attempt {}.'.format(username, i))
                    break
                except Exception:
                    etype, value, tb = sys.exc_info()
                    exc_info = traceback.format_exception(etype, value, tb)
                    printq('  Fail attempt {} at reading {}; {} attempts left'.format(i, login_file, len(parsers) - 1 - i), exc_info)
        else:
            printq("  File {} doesn't exist; can't get login for {} from it.".format(login_file, server))
    return username, password


def get_public_rdb_login(server):
    """
    Returns the public or guest username and password for server
    :param server: string
    :return: (string, string)
    """
    default_user_pass = ('guest', 'guest_pwd')  # Applies to d3drdb, d3dsqlsrvr, d3sqlsrvr, east_database
    return {'huez': ('d3dpub', 'd3dpub')}.get(translate_RDBserver(server), default_user_pass)


def set_rdb_password(server, username=None, password=None, guest=False):
    """
    Sets up an encrypted password for OMFIT to use with SQL databases on a specific server
    :param server: string
        The server this credential applies to
    :param username: string
        Username on server. If a password is specified, this defaults to os.environ['USER'].
        If neither username nor password is specified, OMFIT will try to read both from a login file.
    :param password: string
        The password to be encrypted. Set to '' to erase the exting password for username@server, if there is one.
        If None, OMFIT will attempt to read it from a default login file, like .pgpass. This may or may not be the right
        password for server.
    :param guest: bool
        Use guest login and save it. Each server has its own, with the default being guest // guest_pwd .
    """
    cred = _get_rdb_credential(server)
    if password == '':
        printq('Reset credentials because password was ""')
        return reset_credential(credential=cred)
    elif guest:
        username, password = get_public_rdb_login(server)
    elif password is None:
        username, password = _read_rdb_password_from_file(server)
    elif username is None:
        username = os.environ['USER']

    if username is None or password is None:
        raise ValueError(
            'Failed to figure out a username and password combination to use for server {}. '
            'Try again with guest=True to load public/guest login.'
        )
    userpass = '{} {}'.format(username, password)
    encrypt_credential(credential=cred, password='', otp=userpass)
    printi('Encrypted credentials for {}@{}'.format(username, (translate_RDBserver(server))))
    return


def get_rdb_password(server, username=None, password=None, guest=False, guest_fallback_allowed=True):
    """
    Returns the RDB username and password for server
    :param server: string
        Servers can have different passwords, so you have to tell us which one you're after
    :param username: string
        Defaults to os.environ['USER']
    :param password: string [optional]
        Override decrypted credential and just return this input instead
    :param guest: bool
        Return guest login for server. Each server has its own, with the default being guest // guest_pwd .
    :param guest_fallback_allowed: bool
        Allowed to return guest credentials if look up fails. Otherwise, raise ValueError
    :return: (string, string)
        The username and password for server
    """
    if guest:
        return get_public_rdb_login(server)
    if username is None or password is None:
        userpass = decrypt_credential(_get_rdb_credential(server))[1].split(' ')
        username, password = userpass[0], ' '.join(userpass[1:])
        if not len(password):
            try:
                set_rdb_password(server)
            except Exception:
                pass
            else:
                userpass = decrypt_credential(_get_rdb_credential(server))[1].split(' ')
                username, password = userpass[0], ' '.join(userpass[1:])
    if not len(password):
        if guest_fallback_allowed:
            printq(f'Password lookup failed for {server}. Using guest/public credentials...')
            return get_public_rdb_login(server)
        raise ValueError(
            '''Use set_rdb_password('{server:}') to encrypt your password for {server:} (translates to {t_server:}).
            If your password for {server:} is in .pgpass in your home directory, it will be read from there.
            Otherwise, you must provide it to set_rdb_password().
            Your password for huez would be in .pgpass or .pgdb
            Your password for d3drdb would be in D3DRDB.sybase_login or d3drdb.sybase_login
            Your password for d3dsqlsrvr would be in D3SQLSRVR.sybase_login or d3sqlsrvr.sybase_login
            '''.format(
                server=server, t_server=translate_RDBserver(server)
            )
        )
    else:
        return username, password


class OMFITrdb(SortedDict):
    """
    Class used to connect to relational databases
    """

    def __init__(self, query=None, db='d3drdb', server='d3drdb', by_column=False):
        """
        :param query: string
            SQL SELECT query

        :param db: string
            database to connect to

        :param server: string
            SQL server to connect to (e.g. `d3drdb` as listed under OMFIT['MainSettings']['SERVER'])

        :param by_column: bool
            False: return results by rows (can be slow for large number of records)
                Result of .select() is a SortedDict with keys numbered 0, 1, ... and each holding another SortedDict,
                within which keys correspond to columns
            True: return results by columns
                Result of .select() is a SortedDict with keys corresponding to columns,
                each holding an array with length set by the number of rows selected.
        """
        SortedDict.__init__(self)
        self.dynaLoad = True

        self.query = query
        self.db = db
        self.server = server

        self.cnxn = None
        self.cursor = None

        self.by_column = by_column

    @property
    def postgresql(self):
        return self.db == 'd3d' or self.server == 'huez'

    def _connect(self):
        """Connects to the database"""
        connection = self._setup_connection()
        try:
            if self.postgresql:
                import psycopg2

                try:
                    self.cnxn = psycopg2.connect(**connection)
                except psycopg2.OperationalError:
                    connection = self._setup_connection(force_tunnel=True)
                    self.cnxn = psycopg2.connect(**connection)
            else:
                self.cnxn = pyodbc.connect(**connection)
            self.cursor = self.cnxn.cursor()
        except Exception as _excp1:
            if "Login failed for user " in repr(_excp1) and self.server == 'd3drdb':
                raise ValueError(
                    """
Connection to the D3DRDB relational database requires the file D3DRDB.sybase_login to be present in the $HOME directory of the user on the workstation where OMFIT is running.
You should be able to copy this file from the $HOME folder on the GA iris cluster.
If this file is not available to you contact the GA computer group: iris-support@fusion.gat.com
"""
                )
            connection2 = {}
            try:
                if self.postgresql:
                    connection2 = connection
                    raise
                dsn = 'OMFIT_tmp'
                self._buildUserConfig(connection, dsn)
                connection2['DSN'] = dsn
                connection2['UID'] = connection['UID']
                connection2['PWD'] = connection['PWD']
                self.cnxn = pyodbc.connect(**connection2)
                self.cursor = self.cnxn.cursor()
            except Exception as _excp2:
                printe('\nFailed connection details: ' + repr(_excp1))
                printe(connection)
                printe('\nFailed connection details: ' + repr(_excp2))
                printe(connection2)
                printe('\n')
                raise

    @staticmethod
    def _buildUserConfig(connection, dsn):
        """
        Sets up OMFIT-related controls in the .odbc.ini file in the user's home directory
        :param connection: dict
            Contains connection information. Typically formed by self._setup_connection()
        :param dsn: string
            This is a tag or header name for the relevant section of the .odbc.ini file
        """
        filename = os.environ['HOME'] + os.sep + '.odbc.ini'

        # Create if it did not exist and give OMFIT full control over it
        header = '#--below this line OMFIT is in control of this file--\n'
        if not os.path.exists(filename):
            with open(filename, 'w') as f:
                f.write(header)

        # Retain what is above the header line
        with open(filename, 'r') as _f:
            top = _f.read().split(header)[0]

        # Text to add to .odbc.ini
        txt = []
        c = ''
        if '[' + dsn + ']' in top:
            c = '# '
        txt.append('')
        txt.append(c + '[' + dsn + ']')
        for item in connection:
            txt.append(c + item + ' = ' + str(connection[item]))
        txt.append('')

        # Write updated .odbc.ini
        with open(filename, 'w') as f:
            f.write(top + header + '\n'.join(txt))
        return

    def select(self, query=None, by_column=None):
        """
        Pass a query to the database, presumably with a SELECT statement in it

        :param query: string
            A string such as "SELECT * from profile_runs where shot=126006";
            if None then use the query used to instantiate the object
        :param by_column: bool [optional]
            If True or False, override the self.by_column set at instantiation
        :return: SortedDict
            by_column = False:
                Keys are numbered 0, 1, 2, ...
                The value behind each key is a SortedDict
                The keys of the child SortedDict are the columns selected
            by_column = True:
                Keys are column names
                Values are arrays with length set by the number of rows selected
        """
        if query is None and self.query is None:
            return SortedDict()
        if self.cnxn is None or self.cursor is None:
            self._connect()
            if not self.postgresql:
                self.cursor.execute("use " + self.db)

        if query is None:
            query = self.query
        if by_column is None:
            by_column = self.by_column
        self.cursor.execute(query)
        rows = self.cursor.fetchall()
        if self.postgresql:
            colnames = [desc[0] for desc in self.cursor.description]
        else:
            colnames = None

        res = SortedDict()
        # Not by column can be very computationally expensive if there are many records
        if not by_column:
            for r, row in enumerate(rows):
                res[r] = SortedDict()
                for c, col in enumerate(row):
                    if not self.postgresql:
                        name = row.cursor_description[c][0]
                    else:
                        name = colnames[c]
                    res[r][name] = row[c]

        # Faster, by column organization
        else:
            for r, row in enumerate(rows):
                names = []
                for c, col in enumerate(row):
                    if not self.postgresql:
                        name = row.cursor_description[c][0]
                    else:
                        name = colnames[c]
                    if name in names:
                        continue
                    names.append(name)
                    if r == 0:
                        res[name] = []
                    res[name].append(row[c])
            for name in list(res.keys()):
                res[name] = np.array(res[name])

        return res

    def alter_add(self, table, column_name, column_type, commit=True, verbose=True):
        """
        Alter table in SQL database by adding a column

        :param table: string
            table name

        :param column_name: string
            column name

        :param column_type: string
            column type, has to be an SQL DataType, e.g. BOOLEAN, FLOAT etc..

        :param commit: bool
            commit alter command to SQL

        :param verbose: bool
            print SQL command being used
        """
        if self.postgresql:
            raise NotImplementedError('File issue on Github if this is needed')

        if self.cnxn is None or self.cursor is None:
            self._connect()
            self.cursor.execute("use " + self.db)

        sql_cmd = "ALTER TABLE %s ADD %s %s" % (table, str(column_name), str(column_type))

        if verbose:
            printi(sql_cmd)

        self.cursor.execute(sql_cmd)
        if commit:
            self.commit()
        return sql_cmd

    def alter_drop(self, table, column_name, commit=True, verbose=True):
        """
        Alter table in SQL database by dropping a column

        :param table: string
            table name

        :param column_name: string
            column name

        :param commit: bool
            commit alter command to SQL

        :param verbose: bool
            print SQL command being used

        :return: string
            SQL command
        """
        if self.postgresql:
            raise NotImplementedError('File issue on Github if this is needed')

        if self.cnxn is None or self.cursor is None:
            self._connect()
            self.cursor.execute("use " + self.db)

        sql_cmd = "ALTER TABLE %s DROP COLUMN %s" % (table, str(column_name))

        if verbose:
            printi(sql_cmd)

        self.cursor.execute(sql_cmd)
        if commit:
            self.commit()
        return sql_cmd

    def delete(self, table, where, commit=True, verbose=True):
        """
        Delete row(s) in SQL database

        :param table: string
            table where to update

        :param where: string or dict
            Which record or records should be deleted
            NOTE that all records that satisfy this condition will be deleted!
            A dict will be converted into a string of the form "key1=value1 and key2=value2 ..."

        :param commit: bool
            commit delete to SQL

        :param verbose: bool
            print SQL command being used
        """
        if self.cnxn is None or self.cursor is None:
            self._connect()
            self.cursor.execute("use " + self.db)

        if isinstance(where, dict):
            where = ' and '.join([str(x[0]) + '=' + repr(x[1]) for x in zip(list(where.keys()), list(where.values()))])
        sql_cmd = "DELETE FROM %s WHERE %s" % (table, where)

        if verbose:
            printi(sql_cmd)

        self.cursor.execute(sql_cmd)
        if commit:
            self.commit()
        return sql_cmd

    def commit(self):
        """
        Commit commands in SQL database
        """

        printq('committing')
        self.cnxn.commit()

    def update(self, table, data, where, commit=True, overwrite=1, verbose=True):
        """
        Update row(s) in SQL database

        :param table: string
            Table to update.

        :param data: dict
            Keys are columns to update and values are values to put into those columns.

        :param where: dict or string
            Which record or records should be updated.
            NOTE that all records that satisfy this condition will be updated!
            If it's a dictionary, the columns/data condition will be concatenated with " AND ", so that
                {'my_column': 5.2, 'another_col': 7} becomes "my_column=5.2 AND another_col=7".
            A string will be used directly.

        :param commit: bool
            Commit update to SQL. Set to false for testing without editing the database.

        :param overwrite: bool or int
            0/False: If any of the keys in data already have entries in the table, do not update anything
            1/True: Update everything. Don't even check.
            2: If any of the keys in dat already have entries in the table, don't update those,
               but DO write missing entries.

        :param verbose: bool
            Print SQL command being used.

        :return: string
            The SQL command that would be used.
            If the update is aborted due to overwrite avoidance, the SQL command will be prefixed by "ABORT:"
        """
        if self.cnxn is None or self.cursor is None:
            self._connect()
            if not self.postgresql:
                self.cursor.execute("use " + self.db)

        def mk_kv_list(a):
            """
            Turns a dictionary of columns and values into a list of strings of the form 'k=v'
            to be later concatenated by ", " or " AND "
            :param a: dict-like
            :return: list of strings
            """
            return [str(x[0]) + '=' + x[1] for x in zip(list(a.keys()), tolist(self._sqlRepr(list(a.values()))))]

        if isinstance(where, dict):
            where = ' AND '.join(mk_kv_list(where))
        if len(where):
            where = ' WHERE ' + where
        columns_values = ', '.join(mk_kv_list(data))

        sql_cmd = "UPDATE %s SET %s%s" % (table, columns_values, where)
        if verbose:
            printi(sql_cmd)

        if overwrite not in [1, True, 1.0]:
            # Need to check for pre-existing data
            columns = ', '.join(list(data.keys()))
            query = 'SELECT {columns:} FROM {table:}{where:}'.format(**locals())
            results = self.select(query=query, by_column=False)
            results_with_content = []
            for result in results.values():
                cleaned_result = {k: v for k, v in result.items() if v is not None}
                if len(cleaned_result):
                    results_with_content += [cleaned_result]
            if len(results_with_content) and overwrite == 0:
                printw('Requested update aborted to avoid overwriting some data.')
                return 'ABORT:' + sql_cmd
            elif len(results_with_content) and overwrite > 1:
                skipped_columns = []
                for cleaned_result in results_with_content:
                    for key in cleaned_result:
                        data.pop(key, None)
                        skipped_columns += [key]
                if len(data) == 0:
                    printw('To avoid overwriting, all data were removed from the update list & there is nothing to do.')
                    return 'ABORT:' + sql_cmd
                else:
                    columns_values = ', '.join(mk_kv_list(data))
                    sql_cmd = "UPDATE %s SET %s%s" % (table, columns_values, where)
                    printq('New update list: {}'.format(columns_values))
                    printq('Data removed: {}'.format(list(set(skipped_columns))))
                    if verbose:
                        print('Update command after avoiding overwriting anything:')
                        printi(sql_cmd)

        self.cursor.execute(sql_cmd)
        if commit:
            self.commit()
        return sql_cmd

    def insert(self, table, data, duplicate_key_update=False, commit=True, verbose=True):
        """
        Insert row in SQL database

        :param table: string
            table where data will be inserted

        :param data: dict, list, or tuple
            dict: keys are column names
            list or tuple:
            dictionary (columns & values) or list/tuple of values

        :param duplicate_key_update: bool
            append ' ON DUPLICATE KEY UPDATE' to INSERT command

        :param commit: bool
            commit insert to SQL

        :param verbose: bool
            print SQL command being used

        :return: The SQL command that was used or would be used
        """
        if self.cnxn is None or self.cursor is None:
            self._connect()
            if not self.postgresql:
                self.cursor.execute("use " + self.db)

        if isinstance(data, (list, tuple)):
            values = (','.join(["%s"] * len(data))) % tuple(self._sqlRepr(data))
            sql_cmd = "INSERT INTO %s VALUES (%s)" % (table, values)
        else:
            columns = (','.join(["%s"] * len(data))) % tuple(data.keys())
            values = (','.join(["%s"] * len(data))) % tuple(self._sqlRepr(list(data.values())))
            sql_cmd = "INSERT INTO %s (%s) VALUES (%s)" % (table, columns, values)

        if duplicate_key_update:
            sql_cmd += ' ON DUPLICATE KEY UPDATE'

        if verbose:
            printi(sql_cmd)

        self.cursor.execute(sql_cmd)
        if commit:
            self.commit()
        return sql_cmd

    def copy_row(self, table, **kw):
        r"""
        Copy one row of a table to a new row of a table

        :param table: string
            table in the database for which to copy a row
        :param \**kw: The keywords passed must be the primary keys of the table.
               The values of the keywords must be two-element containers: (copy_from, copy_to)
        :return: string
            SQL command
        """
        if not self.postgresql:
            raise NotImplemented('File issue on GitHub if needed')
        if self.cnxn is None or self.cursor is None:
            self._connect()
            if not self.postgresql:
                self.cursor.execute("use " + self.db)

        primary = self.primary_keys(table)
        for k in kw:
            if k not in primary:
                raise ValueError('%s is not a primary key; only primary keys allowed for copying a row' % k)
        where = []
        where2 = []
        for k, v in list(kw.items()):
            where.append('%s=%s' % (k, self._sqlRepr(v[0])))
            where2.append('%s=%s' % (k, self._sqlRepr(v[1])))
        query = 'SELECT * from %s where %s' % (table, ' and '.join(where))
        copy_from = self.select(query)[0]
        if 'userid' in copy_from:
            copy_from['userid'] = self.username
        for k, v in list(kw.items()):
            copy_from[k] = v[1]
        insert_row = 'INSERT INTO {0}({1}) VALUES ({2});'.format(
            table, ','.join(list(copy_from.keys())), ','.join(['%s'] * len(list(copy_from.values())))
        )
        self.cursor.execute(insert_row, list(copy_from.values()))
        self.commit()
        return insert_row

    def primary_keys(self, table):
        """
        Return the keys that are the primary keys of the table

        :param table: string
            table for which to evaluate the primary keys
        :return: list of strings
        """
        if not self.postgresql:
            raise NotImplemented('File issue on GitHub if needed')
        if self.cnxn is None or self.cursor is None:
            self._connect()
            if not self.postgresql:
                self.cursor.execute("use " + self.db)

        self.cursor.execute("SELECT indexdef from pg_indexes where tablename = '%s'" % table)
        return list(map(lambda x: str(x).strip(), self.cursor.fetchall()[0][0].split('(')[1].split(')')[0].split(',')))

    def _sqlRepr(self, data):
        """
        Forms the SQL representation of data
        :param data: dict
        :return: string
        """
        if self.postgresql:
            # Adapted from Osborne tools setup_run_table._value_to_dbstring:
            # CONVERT PYTHON VALUE TO EXPRESSION FOR ENTRY INTO DB TABLE
            sl = []
            for v in np.atleast_1d(data):
                if isinstance(v, str):
                    s = "'" + v + "'"
                elif v is None:
                    s = 'NULL'
                elif np.iterable(v):
                    s = '{'
                    for x in v:
                        if isinstance(x, str):
                            s = s[:] + '"' + x + '",'
                        else:
                            s = s[:] + str(x) + ','
                    if len(v):
                        s = s[:-1] + '}'
                    else:
                        s = s[:] + '}'
                    s = "'" + s + "'"
                else:
                    s = str(v)
                sl.append(s)
            if len(sl) == 1:
                return sl[0]
            return sl
        else:
            data_out = []
            for k, d in enumerate(np.atleast_1d(data)):
                if d is None or (is_numeric(d) and np.isnan(d)):
                    data_out.append('NULL')
                else:
                    data_out.append(repr(d))
            return data_out

    def __getstate__(self):
        return {
            'query': self.query,
            'db': self.db,
            'server': self.server,
            'cnxn': None,
            'cursor': None,
            'dynaLoad': self.dynaLoad,
            'by_column': self.by_column,
        }

    def __setstate__(self, new_state):
        """
        :param new_state: dict or tuple containing dict in [0]
        """
        if isinstance(new_state, tuple):
            new_state = new_state[0]
        for k in list(new_state.keys()):
            setattr(self, k, new_state[k])
        if 'by_column' not in new_state:
            self.by_column = False
        return

    @dynaLoad
    def load(self):
        """
        Connect to the database and retrieve its content
        """
        if self.query is not None:
            self.clear()
            res = self.select()
            SortedDict.update(self, res)

    def _setup_connection(self, force_tunnel=False):
        # Get server/tunnel from OMFIT['MainSettings']['SERVER'] list
        server0 = translate_RDBserver(self.server, _RDBserverDict)
        server = OMFIT['MainSettings']['SERVER'][server0]['RDB_server']
        tunnel = OMFIT['MainSettings']['SERVER'][server0]['tunnel']

        # Try directly connecting to the RDB server
        if server not in OMFITaux['RDBserverReachable']:
            OMFITaux['RDBserverReachable'][server] = test_connection(None, *parse_server(server)[2:], timeout=1, ntries=1)

        rdb_server = list(parse_server(server))
        username, password = get_rdb_password(server0)
        if username != rdb_server[0]:
            printw(
                "username from get_rdb_password({}) ({}) != username indicated by MainSettings ({})".format(
                    server0, username, rdb_server[0]
                )
            )
        self.username = username

        # If direct connection fails, use tunneling
        if not OMFITaux['RDBserverReachable'][server] or force_tunnel:
            # Handle server tunneling (this function does already buffering)
            rdb_server[2:] = list(map(str, setup_ssh_tunnel(server, tunnel, True)))[1:]

        connection = {}
        if not self.postgresql:
            connection['UID'] = username
            connection['PWD'] = password
            connection['SERVER'] = rdb_server[2]
            connection['PORT'] = rdb_server[3]
            driver = 'tdsodbc'
            if 'driver' in list(OMFIT['MainSettings']['SERVER'][server0].keys()):
                connection['DRIVER'] = driver = OMFIT['MainSettings']['SERVER'][server0]['driver']
            if isinstance(driver, str) and driver and os.sep not in driver:
                connection['DRIVER'] = find_library(driver)
            if not isinstance(connection['DRIVER'], str) or not connection['DRIVER']:
                raise OMFITexception(
                    'The library to drive your `' + str(driver) + '` SQL connection is not setup properly.\n' + repr(connection)
                )
            if 'tdsodbc' in driver:
                connection['TDS_Version'] = '8.0'
            for k in list(OMFIT['MainSettings']['SERVER'][server0].keys()):
                if k not in list(connection.keys()) + ['driver', 'server', 'tunnel', 'RDB_server']:
                    connection[k] = OMFIT['MainSettings']['SERVER'][server0][k]
        else:
            connection['user'] = username
            connection['password'] = password
            connection['host'] = rdb_server[2]
            connection['port'] = rdb_server[3]
            connection['database'] = self.db
        printq('')
        printq(server)
        for k in list(connection.keys()):
            printq(k + '\t\t:\t ' + str(connection[k]))

        return connection


def available_EFITs(scratch, device, shot, default_snap_list=None, format='{tree}'):
    """
    :param scratch: dict
        dictionary where the information is cached (to avoid querying SQL database multiple times)

    :param device: string
        device for which to query list of available EFITs

    :param shot: int
        shot for which to query list of available EFITs

    :param default_snap_list: dict
        dictionary to which list of available efits will be passed

    :param format: string
        format in which to write list of available efits (tree, by, com, drun, runid) are possible options

    :return: dict
        dictionary with list of available options formatted as {text:format}
    """
    efit_sql = 'EFIT_SQL_%s' % str(shot)
    if default_snap_list is None:
        default_snap_list = {'': ''}
    snap_list = copy.deepcopy(default_snap_list)
    help = ''
    if is_device(device, 'DIII-D'):
        try:
            query = "SELECT * FROM plasmas WHERE shot=" + str(shot) + " AND code_name='EFIT' AND deleted!=1 ORDER BY tree"
            try:
                if scratch[efit_sql].query != query:
                    scratch[efit_sql] = OMFITrdb(query, 'code_rundb')
            except Exception:
                scratch[efit_sql] = OMFITrdb(query, 'code_rundb')
            runids = scratch[efit_sql].across("['*']['shot']")
            try:
                runids = scratch[efit_sql].across("['*']['run_id']")
            except Exception:
                pass
            for tree, by, com, drun, tag, runid in zip(
                *(
                    scratch[efit_sql].across("['*']['tree']"),
                    scratch[efit_sql].across("['*']['run_by']"),
                    scratch[efit_sql].across("['*']['run_comment']"),
                    scratch[efit_sql].across("['*']['date_run']"),
                    scratch[efit_sql].across("['*']['runtag']"),
                    runids,
                )
            ):
                if tree == 'EFIT':
                    item = by + ' ' + tag
                    tree = int(runid)
                elif by in ['autoload', 'mdsadmin']:
                    item = f'{tree} {tag}'
                else:
                    item = f'{tree} by {tag}'

                help += '[' + item + '] : "' + com.strip() + '"\n'
                if drun is not None:
                    snap_list['[' + item + '] ' + drun.strftime(' (%d %b %Y @ %H:%M)')] = format.format(
                        tree=tree, runid=runid, by=by, com=com, drun=drun, tag=tag
                    )

                else:
                    snap_list['[' + item + '] ' + ' (???)'] = format.format(tree=tree, runid=runid, by=by, com=com, drun=drun, tag=tag)

        except Exception as _excp:
            raise
            printe('Could not retrieve list of available EFIT runs: ' + repr(_excp))
            snap_list = {'EFIT01': 'EFIT01', 'EFIT02': 'EFIT02'}
            help = 'Enter EFIT tree by hand\n\nCould not retrieve list of available EFIT runs:\n' + repr(_excp)
    help = help.strip()
    return snap_list, help



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii
from omfit_classes.omfit_nc import OMFITnc
from omfit_classes import namelist

from collections import OrderedDict

__all__ = ['OMFIToedgeInput', 'OMFIToedgeRun', 'OMFIToedgeNC']


def interpret(x):

    x = x.split('\'')
    for k in range(1, len(x), 1)[::2]:
        x[k] = '\'' + x[k] + '\''
    x = [_f for _f in x if _f]

    xx = []
    for k in x:
        if k.startswith('\''):
            xx.append(k)
        else:
            xx.extend(list(map(namelist.interpreter, [_f for _f in re.split(' |\t', k) if _f])))

    data = []
    inline_comment = ''

    for k in range(len(xx)):
        if not len(inline_comment) and is_numeric(xx[k]):
            data.append(xx[k])
        elif not len(inline_comment) and xx[k][0] == '\'':
            data.append(xx[k][1:-1])
        else:
            inline_comment += ' ' + str(xx[k])
    inline_comment = inline_comment.strip(' \'')

    return data, inline_comment


class OMFIToedgeInput(SortedDict, OMFITascii):
    """OEDGE input data file"""

    def __init__(self, filename, casename=None, caseloc=None, root=None, **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self)

        self.dynaLoad = True

        printd('Init OEDGE input file: ', casename, caseloc)

        self.casename = casename
        self.caseloc = caseloc

        # set casename and treelocation of the object
        self.setloc(casename, caseloc, root)

    def __call__(self, *args):
        # Allow the case object to return the string representation of its tree elements
        # by checking case is eval(case()) in the calling routine - can verify object is loaded where expected
        if len(args) == 0:
            return self.caseloc
        elif self.caseloc is not None:
            tmp = self.caseloc
            for arg in args:
                tmp = tmp + "['" + arg + "']"
            return tmp

    def setloc(self, casename=None, caseloc=None, root=None):
        # Rotuine will assign the casename and location in the tree either to passed in values
        # or it will attempt to determine them based on the filename and treelocation return values
        # print('filename:',self.filename)
        # print('location:',treeLocation(self))
        # print('inputs:',casename,caseloc)
        # print('current:',self.casename,self.caseloc)

        # Allow overwrite of case name  ... do I need to update filename and write new file (?)
        # leave as a separate case in the IF statement so this can be addressed later
        if self.casename is None and casename is None:
            self.casename, filetype = os.path.splitext(self.filename.split(os.sep)[-1])
        elif self.casename is None and casename is not None:
            self.casename = casename
        elif casename is not None:
            self.casename = casename

        # Case is not in the tree when first created on module load .. object is added to the
        # tree after it is created (of course) so treeLocation is not usable until later.

        if self.caseloc is None and caseloc is None:
            #  check correct location in tree first ... then use treeLocation if OMFIT not available and root not specified
            if self.casename is not None:
                try:
                    # try is needed because dictionary entry may not be accessible at module load
                    # If root is not specified then try using the absolute path
                    if root is None:
                        if self.casename in OMFIT['OEDGE']['INPUTS']['CASES']:
                            # check to see if the object is loaded into the tree in the expected location
                            if self is OMFIT['OEDGE']['INPUTS']['CASES'][self.casename]:
                                self.caseloc = "OMFIT['OEDGE']['INPUTS']['CASES'][%s]" % repr(self.casename)
                            else:
                                self.caseloc = None
                    else:
                        if self.casename in root['INPUTS']['CASES']:
                            # check to see if the object is loaded into the tree in the expected location
                            if self is root['INPUTS']['CASES'][self.casename]:
                                self.caseloc = "root['INPUTS']['CASES'][%s]" % repr(self.casename)
                            else:
                                self.caseloc = None

                except KeyError:
                    self.caseloc = treeLocation(self)[-1]
                    # when the object is not loaded in the tree then treeLocation returns the null string
                    # replace with None
                    if self.caseloc == '':
                        self.caseloc = None

        elif self.caseloc is None and caseloc is not None:
            self.caseloc = caseloc
        elif caseloc is not None:
            self.caseloc = caseloc

    @dynaLoad
    def load(self):

        # empty the dictionary in case this is a reload
        self.clear()

        with open(self.filename, 'r') as f:
            lines = f.read()

        # handle inline comments with $
        lines = re.sub(r'(?<!\n)\$', '\n$', lines)
        lines = lines.split('\n')

        # split sections
        c = 0
        inhibit = False
        for line0 in lines:
            line = line0.rstrip()
            # stop parsing when find a line starting with c
            if line.startswith('c'):
                inhibit = True
                self['__comment_%06d__' % c] = line0
                c += 1
            # new entry with Z'{ style
            elif not inhibit and line.startswith(r'\'{'):
                item = re.sub(r'.*(\{.*\}).*', r'\1', line.split("'")[1])
                self[item] = SortedDict()
                self[item]['__raw__'] = [line]
            # new entry with '* or '+ style
            elif not inhibit and re.match(r'^\'[\*\+][a-zA-Z0-9][0-9][0-9] ', line):
                item = line.split("'")[1][:4]
                # Item C14 is repeated twice in some input files - the second should be C15
                if item in self:
                    # found a repeated item
                    # print("Repeat item found:",item)
                    # if the repeat item is C14 replace with C15
                    if item == '+C14':
                        item == '+C15'
                self[item] = SortedDict()
                self[item]['__raw__'] = [line]
            # continuation of entry
            elif not inhibit and re.match(r'^ *[\'0-9\.\-].*', line):
                # Think this is where the other mislabeled lines go
                #
                # Kludgy fix to a bug in older input files
                #
                if 'Ring Range :-' in line:
                    # insert tag at beginning of line
                    item = '+F07'
                    line = "'" + item + line[1:]
                    self[item] = SortedDict()
                    self[item]['__raw__'] = [line]
                elif 'J0 & J1    :-' in line:
                    item = '+F08'
                    line = "'" + item + line[1:]
                    self[item] = SortedDict()
                    self[item]['__raw__'] = [line]
                elif 'VMF0,1,2   :-' in line:
                    item = '+F09'
                    line = "'" + item + line[1:]
                    self[item] = SortedDict()
                    self[item]['__raw__'] = [line]
                else:
                    self[item]['__raw__'].append(line)
            # comment
            else:
                self['__comment_%06d__' % c] = line0
                c += 1

        # interpret sections
        for item in list(self.keys()):
            if item.startswith('__comment_'):
                continue

            # initialize the tag field for all non-comment input items
            self[item]['tag'] = ''

            raw = self[item]['__raw__']
            raw0 = raw[0].strip().split('\'')
            # special {...}
            if '{' in item:
                self[item]['comment'] = raw0[1].split('}')[1].strip()
                self[item]['data'] = []
                self[item]['tag'] = item
                self[item]['inline_comment'] = ''

                try:
                    # scalar
                    if not raw0[2]:
                        raise
                    self[item]['data'], inline_comment = interpret(raw0[2])
                    # scalars as scalars
                    if len(self[item]['data']) == 1:
                        self[item]['data'] = self[item]['data'][0]

                except Exception:
                    # array
                    for line in raw[1:]:
                        data, inline_comment = interpret(line)
                        self[item]['data'].append(data)
            # standard
            else:
                self[item]['comment'] = raw0[1][5:].strip()
                self[item]['tag'] = item[1:]

                # scalar(ish)
                if len(raw) == 1:
                    tmp = interpret(raw[0])
                    self[item]['data'] = tmp[0][1:]
                    self[item]['inline_comment'] = tmp[-1]
                    # scalars as scalars
                    if len(self[item]['data']) == 1:
                        self[item]['data'] = self[item]['data'][0]

                # array
                else:
                    self[item]['inline_comment'] = '\''.join(raw0[2:])
                    self[item]['inline_comment'] = self[item]['inline_comment'].strip(' \'')
                    for ndummies, line in enumerate(raw[1:]):
                        if "' '" not in line:
                            break
                    header = re.sub(r'(\'.*\'\s*)([0-9]+)(\s.*)', r'\1', raw[1 + ndummies] + ' ').strip('\' ')
                    n = int(re.sub(r'(\'.*\'\s*)([0-9]+)(\s.*)', r'\2', raw[1 + ndummies] + ' ').strip('\' '))
                    header_comment = re.sub(r'(\'.*\'\s*)([0-9]+)(\s.*)', r'\3', raw[1 + ndummies] + ' ').strip()
                    data = []
                    for k in range(len(raw[2 + ndummies :])):
                        data.append(interpret(raw[2 + ndummies + k])[0])
                    dummies = raw[1 : ndummies + 1]
                    self[item]['dummies'] = dummies
                    self[item]['header'] = header
                    self[item]['n'] = n
                    self[item]['header_comment'] = header_comment
                    self[item]['data'] = data

    def create_output(self):
        # set line lengths to give a more nicely formatted output file that is more legible
        # data for some inputs in OEDGE needs to be within column 72
        # in-line comments can be beyond the end of the line
        tag_width = 4
        comment_width = 50
        end_width = 20
        data_width = 10

        # create output list
        out = []
        for item in self:
            if item.startswith('__comment_'):
                out.append(self[item].strip().ljust(comment_width))

            elif '{' in item:
                out.append('\'%s %s\'' % (item.ljust(tag_width), self[item]['comment'].strip().ljust(comment_width)))
                if isinstance(self[item]['data'], list):
                    for line in self[item]['data']:
                        out.append(' '.join(item.rjust(data_width) for item in map(repr, line)))
                else:
                    out[-1] += ' ' + str(self[item]['data']).strip().rjust(data_width)

            elif 'header_comment' in self[item]:
                if self[item]['comment'] == '':
                    out.append(
                        '\'%s %s\' \'%s\'' % (item.ljust(tag_width), self[item]['comment'].strip(), self[item]['inline_comment'].strip())
                    )
                    for dummy in self[item]['dummies']:
                        out.append(dummy.strip().ljust(comment_width))
                    out.append(
                        '\'%s\' %s %s'
                        % (
                            self[item]['header'].strip().ljust(comment_width),
                            str(len(self[item]['data'])).rjust(data_width),
                            self[item]['header_comment'].rjust(end_width),
                        )
                    )
                    for line in self[item]['data']:
                        out.append(' '.join(item.rjust(data_width) for item in map(repr, line)))
                else:
                    out.append(
                        '\'%s %s\'  %s ' % (item.ljust(tag_width), self[item]['comment'].strip(), self[item]['inline_comment'].strip())
                    )
                    for dummy in self[item]['dummies']:
                        out.append(dummy.strip().ljust(comment_width))
                    out.append(
                        '\'%s\' %s %s'
                        % (
                            self[item]['header'].strip().ljust(comment_width),
                            str(len(self[item]['data'])).rjust(data_width),
                            self[item]['header_comment'].rjust(end_width),
                        )
                    )
                    for line in self[item]['data']:
                        out.append(' '.join(item.rjust(data_width) for item in map(repr, line)))

            elif 'inline_comment' in self[item]:
                if self[item]['comment'] == '':
                    out.append(
                        '\'%s %s\' %s %s'
                        % (
                            item.ljust(tag_width),
                            self[item]['comment'].strip(),
                            ' '.join(item.rjust(data_width) for item in map(repr, tolist(self[item]['data']))),
                            self[item]['inline_comment'],
                        )
                    )
                elif isinstance(self[item]['data'], str):
                    out.append(
                        '\'%s %s\' %s %s'
                        % (
                            item.ljust(tag_width),
                            self[item]['comment'].strip(),
                            ' '.join(item.rjust(data_width) for item in map(repr, tolist(self[item]['data']))),
                            self[item]['inline_comment'],
                        )
                    )
                else:
                    out.append(
                        '\'%s %s\' %s %s'
                        % (
                            item.ljust(tag_width),
                            self[item]['comment'].strip().ljust(comment_width),
                            ' '.join(item.rjust(data_width) for item in map(repr, tolist(self[item]['data']))),
                            self[item]['inline_comment'].rjust(end_width),
                        )
                    )

        return out

    # Use the OMFITObject base class method "saveas" to save using a different file name
    def altsave(self, filename=None):

        if filename is None:
            filename = self.filename

        out = self.create_output()

        with open(filename, 'w') as f:
            f.write('\n'.join(out))

    @dynaSave
    # Use the OMFITObject base class method "saveas" to save using a different file name
    def save(self, **kw):

        out = self.create_output()

        with open(self.filename, 'w') as f:
            f.write('\n'.join(out))


class OMFIToedgeRun(SortedDict):
    pass

    def __init__(self, casename=None, caseloc=None, runloc=None, plot_file_name=None, grid=None, background=None, fc=None, divref=None):
        self.casename = casename
        self.caseloc = caseloc
        self.runloc = runloc
        self.basestr = ''
        self['casename'] = self.casename
        self['plot'] = self.plot_file_name
        self['grid'] = self.casename
        self['casename'] = self.casename

    def __call__(self, *args):
        # Allow the run object to return the string representation of its tree elements
        # by checking case is eval(case()) in the calling routine - can verify object is loaded where expected
        if len(args) == 0:
            return self.basestr
        else:
            tmp = self.basestr
            for arg in args:
                tmp = tmp + "['" + arg + "']"
            return tmp


class OMFIToedgeNC(OMFITnc):
    """OEDGE output NC data file"""

    plot_types_available = ['contour', 'along ring']
    # possible future plots to be added
    # plot_types_available = ['contour','along ring','diagnostic','along wall','along target']

    # background plasma quantities
    # present for cases with a background plasma (basically everything)
    bg_plots = OrderedDict()

    bg_plots['ne'] = {
        'data': ['KNBS'],
        'targ': ['KNDS'],
        'label': 'Density',
        'units': 'm-3',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    bg_plots['Te'] = {
        'data': ['KTEBS'],
        'targ': ['KTEDS'],
        'label': 'Te',
        'units': 'eV',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    bg_plots['Ti'] = {
        'data': ['KTIBS'],
        'targ': ['KTIDS'],
        'label': 'Ti',
        'units': 'eV',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    bg_plots['Vb'] = {
        'data': ['KVHS'],
        'targ': ['KVDS'],
        'label': 'Vpara',
        'units': 'm/s',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
        'scale': '1/QTIM',
    }
    bg_plots['Plasma'] = {
        'data': ['ne', 'Te', 'Ti', 'Vb'],
        'label': 'Background Plasma',
        'lineaxes': ['P', 'S'],
        'group': True,
        'notebook': False,
    }

    bg_plots['E'] = {
        'data': ['KES'],
        'targ': ['KEDS'],
        'label': 'Epara',
        'units': 'V/m',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    bg_plots['ExB Pol'] = {
        'data': ['E_POL'],
        'targ': None,
        'label': 'Epol',
        'units': 'V/m',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    bg_plots['ExB Rad'] = {
        'data': ['E_RAD'],
        'targ': None,
        'label': 'Epol',
        'units': 'V/m',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    bg_plots['Efields'] = {
        'data': ['E', 'ExB Pol', 'ExB Rad'],
        'label': 'Electric Fields',
        'lineaxes': ['P', 'S'],
        'group': True,
        'notebook': False,
    }

    bg_plots['V_pol'] = {
        'data': ['EXB_P'],
        'targ': None,
        'label': 'Vpol',
        'units': 'm/s',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    bg_plots['V_rad'] = {
        'data': ['EXB_R'],
        'targ': None,
        'label': 'Vrad',
        'units': 'm/s',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    bg_plots['Drifts'] = {'data': ['V_pol', 'V_rad'], 'label': 'Drifts', 'lineaxes': ['P', 'S'], 'group': True, 'notebook': False}

    # Data present for cases that ran EIRENE to generate hydrogenic quantities
    h_plots = OrderedDict()
    h_plots['H Power'] = {
        'data': ['HPOWLS'],
        'targ': None,
        'label': 'Hydrogen Radiated Power',
        'units': 'W/m3',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    h_plots['H Line'] = {
        'data': ['HLINES'],
        'targ': None,
        'label': 'Hydrogen Line Radiation',
        'units': 'W/m3',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    h_plots['H Dalpha'] = {
        'data': ['PINALP'],
        'targ': None,
        'label': 'Hydrogen Dalpha Emission',
        'units': 'ph/m3/s',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    h_plots['H Ionization'] = {
        'data': ['PINION'],
        'targ': None,
        'label': 'Hydrogen Ionization',
        'units': '/m3/s',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    h_plots['H Recomb'] = {
        'data': ['PINREC'],
        'targ': None,
        'label': 'Hydrogen Recombination',
        'units': '/m3/s',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    h_plots['H2 Density'] = {
        'data': ['PINMOL'],
        'targ': None,
        'label': 'Hydrogen Molecule Density',
        'units': '/m3',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    h_plots['H0 Density'] = {
        'data': ['PINATO'],
        'targ': None,
        'label': 'Hydrogen Atom Density',
        'units': '/m3',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    h_plots['H Atom Temp'] = {
        'data': ['PINENA'],
        'targ': None,
        'label': 'Hydrogen Atom Temperature',
        'units': 'eV',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    h_plots['H Mol Temp'] = {
        'data': ['PINENM'],
        'targ': None,
        'label': 'Hydrogen Molecule Temperature',
        'units': 'eV',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    h_plots['H Ion Energy Loss'] = {
        'data': ['PINQI'],
        'targ': None,
        'label': 'Hydrogen-Ion Energy Loss Term',
        'units': 'W/m3(?)',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    h_plots['H Elec Energy Loss'] = {
        'data': ['PINQE'],
        'targ': None,
        'label': 'Hydrogen-Electron Energy Loss Term',
        'units': 'W/m3(?)',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    h_plots['H Quantities'] = {
        'data': ['H0 Density', 'H2 Density', 'H Ionization', 'H Recomb'],
        'label': 'Hydrogen Quantities',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }

    # Data present for cases that ran impurities
    imp_plots = OrderedDict()
    imp_plots['Imp Density'] = {
        'data': ['DDLIMS'],
        'targ': None,
        'label': 'Impurity Density',
        'units': '/m3',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
        'scale': 'ABSFAC',
    }
    imp_plots['Imp Temperature'] = {
        'data': ['DDTS'],
        'targ': None,
        'label': 'Impurity Temperature',
        'units': 'eV',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
    }
    imp_plots['Imp Ionization'] = {
        'data': ['TIZS'],
        'targ': None,
        'label': 'Impurity Ionization',
        'units': '/m3/s',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
        'scale': 'ABSFAC',
    }
    imp_plots['Imp Radiated Power'] = {
        'data': ['POWLS'],
        'targ': None,
        'label': 'Impurity Radiated Power',
        'units': 'W/m3',
        'lineaxes': ['P', 'S'],
        'group': False,
        'notebook': False,
        'scale': 'ABSFAC',
    }
    imp_plots['Impurity Quantities'] = {
        'data': ['Imp Density', 'Imp Temperature', 'Imp Ionization', 'Imp Radiated Power'],
        'label': 'Impurity Quantities',
        'lineaxes': ['P', 'S'],
        'group': True,
        'notebook': False,
    }

    # Plots along surfaces (?) separate category or make a different dict entry under appropriate sections?
    surface_plots = OrderedDict()

    # Synthetic diagnostic plots are a separate category since they will require special treatment
    diagnostic_plots = OrderedDict()

    # available plot types - contour/false colour plots, line plots along ring, line plots along surfaces (use different plotting axes)
    plot_types = ['contour-polygon', 'contour-interpolate', 'along ring', 'surface']
    # default_contour = 'contour-polygon'

    all_plots = OrderedDict()

    all_plots.update(bg_plots)
    all_plots.update(h_plots)
    all_plots.update(imp_plots)
    all_plots.update(surface_plots)

    plot_categories = ['background', 'hydrogen', 'impurity', 'surface', 'diagnostic']

    def __init__(self, filename, **kw):
        OMFITnc.__init__(self, filename, **kw)
        self.dynaLoad = True

    @dynaLoad
    def load(self):

        # empty the dictionary in case this is a reload
        self.clear()

        # print('Load netcdf:')

        OMFITnc.load(self)

        #
        # Do some post-processing after loading
        #
        # add anything else it should automatically do with the data on load ...
        # should these be done in the load routine or in the constructor?
        # Presumably the constructor calls load()

        # print('Load self variables:')
        self.load_simulation_data()
        self.plots_available()

    def load_simulation_data(self):
        # This routine labels specific geometry data from the NC file into local variables
        # This data is frequently used in the plots .. since it appears that one case object
        # handles all the plotting then it is worthwhile adding some efficiencies

        # Certain data must be present to make plots possible ... if it is not available for
        # some reason then flag it as an error.

        # load wall and create path object to define inside and outside the vessel
        # This is used in the tricontourf plots to eliminate triangles outside the vessel
        debug = True

        # vessel wall definition
        self.nves = self['NVES']['data']
        self.rves = self['RVES']['data'][: self.nves]
        self.zves = self['ZVES']['data'][: self.nves]
        self.wall = [[self.rves[i], self.zves[i]] for i in range(self.nves)]
        self.wall_path = matplotlib.path.Path(self.wall)

        # load the computational mesh polygon data into a PolyCollection for plotting code results
        # remove zero volume polygons from the list and also exclude when loading simulation data

        self.nvertp = self['NVERTP']['data']
        self.rvertp = self['RVERTP']['data']
        self.zvertp = self['ZVERTP']['data']
        self.korpg = self['KORPG']['data']

        # cell centers
        self.rs = self['RS']['data']
        self.zs = self['ZS']['data']

        # Significant surfaces in the computational mesh
        self.nrs = self['NRS']['data']
        self.nks = self['NKS']['data']
        self.irsep = self['IRSEP']['data']
        # no irsep2 in netcdf
        # self.irsep2 = self['IRSEP2']['data']
        self.irwall = self['IRWALL']['data']
        self.irwall2 = self['IRWALL2']['data']
        self.irtrap = self['IRTRAP']['data']
        self.irtrap2 = self['IRTRAP2']['data']

        # grid connection map data
        self.ikins = self['IKINS']['data']
        self.ikouts = self['IKOUTS']['data']
        self.irins = self['IRINS']['data']
        self.irouts = self['IROUTS']['data']

        # properties of the grid and target
        self.area = self['KAREAS']['data']

        # idds is an index array mapping from (targ,ir) -> (target index)
        # However, the index is from fortran arrays and numbers from 1
        # so when mapped to python arrays you need to subtract 1 which
        # should work if idds is a np array
        self.idds = self['IDDS']['data'].copy() - 1
        self.smax = self['KSMAXS']['data']
        self.pmax = self['KPMAXS']['data']
        self.kss = self['KSS']['data']
        self.kps = self['KPS']['data']

        mesh = []
        count = 0
        debug = True

        for ir in range(self.nrs):
            for ik in range(self.nks[ir]):
                index = self.korpg[ir, ik] - 1
                if self.area[ir, ik] != 0.0:
                    mesh.append(list(zip(self.rvertp[index][0:4], self.zvertp[index][0:4])))
                    count = count + 1

                # check to see that the cell center point is inside the set of vertices for the cell
                # this verifies that the grid is loading correctly
                if debug and self.area[ir, ik] != 0.0:
                    vert = list(zip(self.rvertp[index][0:4], self.zvertp[index][0:4]))
                    cell = matplotlib.path.Path(vert)
                    r = self['RS']['data'][ir, ik]
                    z = self['ZS']['data'][ir, ik]
                    if not cell.contains_point([r, z]):
                        print("grid loading error:")
                        print("vert:", vert)
                        print("test:", ir, ik, r, z, cell.contains_point([r, z]))

        # save the number of non-zero volume cells in the mesh
        # save the mesh polygons as in grid
        self.num_cells = count
        self.grid = mesh

        # The grid is used to create a polygon plot
        # coll = PolyCollection(self.grid,array=data,cmap=pyplot.cm.jet,edgecolors='none')

        #
        # The above created a mesh of polygons from the computational grid
        # Also needed is a mesh composed of the cell center points both with and without the
        # target coordinates. These meshes have two uses:
        # 1) Interpolated triangular meshes which are based on cell center and target data
        # 2) Interpolation of the OEDGE results to generate an ERO plasma file.
        #

        self.num_points = self.num_cells + (self.nrs - self.irsep + 1) * 2

        self.r_cells = np.zeros(self.num_cells)
        self.z_cells = np.zeros(self.num_cells)

        self.r_points = np.zeros(self.num_points)
        self.z_points = np.zeros(self.num_points)

        cp = 0
        cc = 0

        for ir in range(self.nrs):
            for ik in range(self.nks[ir]):

                if ik == 0 and ir >= self.irsep - 1:
                    # add first target point
                    id0 = self.idds[1, ir]
                    self.r_points[cp] = self['RP']['data'][id0]
                    self.z_points[cp] = self['ZP']['data'][id0]
                    cp = cp + 1

                if self.area[ir, ik] != 0.0:
                    self.r_points[cp] = self.rs[ir][ik]
                    self.z_points[cp] = self.zs[ir][ik]
                    cp = cp + 1
                    self.r_cells[cc] = self.rs[ir][ik]
                    self.z_cells[cc] = self.zs[ir][ik]
                    cc = cc + 1

                if ik == self.nks[ir] - 1 and ir >= self.irsep - 1:
                    # add last target point
                    id1 = self.idds[0, ir]
                    self.r_points[cp] = self['RP']['data'][id1]
                    self.z_points[cp] = self['ZP']['data'][id1]
                    cp = cp + 1

        # Impurity scaling factor if available
        # All impurity results are stored scaled to 1 part/m-tor/s entering the system
        # to get absolute values the results are multipled by absfac (absolute scaling factor)
        if 'ABSFAC' in self:
            self.absfac = self['ABSFAC']['data']
        else:
            self.absfac = 1.0

        # Simulation time steps
        # ion time step
        if 'QTIM' in self:
            self.qtim = self['QTIM']['data']
        else:
            self.qtim = 1.0

        # neutral time step
        if 'FSRATE' in self:
            self.fsrate = self['FSRATE']['data']
        else:
            self.fsrate = 1.0

        # print('qtim:',self.qtim,self.fsrate)

        if 'NIZS' in self:
            self.nizs = self['NIZS']['data']
        else:
            self.nizs = None

    def plots_available(self):
        # this routine returns a list of quantities that this object can plot
        # background plasma plots
        self.plots_avail = []
        self.bg_plots_avail = []
        self.h_plots_avail = []
        self.imp_plots_avail = []
        self.surf_plots_avail = []

        for plotkey in list(OMFIToedgeNC.all_plots.keys()):
            # Check to see that all the required datasets are available in the NC file for each class of plots
            avail = True
            for datakey in OMFIToedgeNC.all_plots[plotkey]['data']:
                dataname, targname, label, scalef = self.get_names(plotkey, datakey)
                if dataname not in self:
                    avail = False
                if targname is not None:
                    if targname not in self:
                        avail = False

            # if all elements of a group plot are available then that plot is available
            if avail:
                self.plots_avail.append(plotkey)
                if plotkey in OMFIToedgeNC.bg_plots:
                    self.bg_plots_avail.append(plotkey)
                elif plotkey in OMFIToedgeNC.h_plots:
                    self.h_plots_avail.append(plotkey)
                elif plotkey in OMFIToedgeNC.imp_plots:
                    self.imp_plots_avail.append(plotkey)
                elif plotkey in OMFIToedgeNC.surface_plots:
                    self.surf_plots_avail.append(plotkey)

    def get_plot_categories(self):
        # return a list of the available plot categories
        return self.plot_categories

    def get_plots_available(self, kind=None):
        # return a list of the plots that are available in the current dataset
        # Since the plot list is getting long - allow for splitting by type so that the UI
        # can have more concise drop down lists
        # print("get_plots_available:",kind)
        if kind is None:
            return self.plots_avail
        elif kind == 'background':
            return self.bg_plots_avail
        elif kind == 'hydrogen':
            return self.h_plots_avail
        elif kind == 'impurity':
            return self.imp_plots_avail
        else:
            return []

    def get_plot_types(self):
        # this routine returns a list of the supported plot types
        return self.plot_types

    def get_along_ring_axis_types(self, plot_select):
        # look up the selected plot and return the along ring or linear axis types available
        if plot_select in OMFIToedgeNC.all_plots:
            return OMFIToedgeNC.all_plots[plot_select]['lineaxes']
        # if this fails for some reason then return None ... but it shouldn't
        return None

    def need_ionization_state(self, selection):
        # Check to see if a charge state is needed for the specified plot
        res = False
        if selection in OMFIToedgeNC.imp_plots:
            res = True
        return res

    def need_ring_number(self, selection):
        # Check to see if a ring number is needed for specified plot type
        res = False
        if selection == 'along ring':
            res = True
        return res

    def get_data_2d(self, dataname, targname=None, charge=None, scalef=1.0):
        # load a set of 2D data
        #

        rawdata = self[dataname]['data']
        count = 0
        #
        # Note: targname should always be None for cases where charge is specified
        # charge is a leading index into the data array mapping.

        # if targname is specified then add the target data at the ends of the rings
        if targname is None:
            # leave out target data when plotting a polygon mesh
            data = np.zeros(self.num_cells)
        else:
            # add space to include target values in the mesh - used in triangulation interpolation plots
            data = np.zeros(self.num_points)

        # loop through data arrays
        # Note ... all of this is required because the rows of the array are of inconsistent lengths .. it isn't a rectangular grid
        # in terms of indexing ... core, sol and pfz rings may all have different numbers of cells though nsol_cells=ncore+npfz is usually true
        # along a ring
        count = 0
        for ir in range(self.nrs):
            for ik in range(self.nks[ir]):
                if targname is not None and ik == 0 and ir >= self.irsep - 1:
                    # add first target point
                    id0 = self.idds[1, ir]
                    data[count] = self[targname]['data'][id0]
                    count = count + 1

                if self.area[ir, ik] != 0.0:
                    # include index for charge state if specified - this requires an offset
                    if charge is None:
                        data[count] = rawdata[ir][ik] * scalef
                    else:
                        # Note: the results array indexing goes from -1:nizs for charge states.
                        # -1 maps to 0 in the python array ... charge 0 is in index 1 - so add one to charge
                        # to get correct indexing
                        data[count] = rawdata[charge + 1][ir][ik] * scalef

                    count = count + 1

                if targname is not None and ik == self.nks[ir] - 1 and ir >= self.irsep - 1:
                    # add last target point
                    id1 = self.idds[0, ir]
                    data[count] = self[targname]['data'][id1]
                    count = count + 1

        return data

    def plot(self, plot_select, plot_type, axis_type='S', ring_range=[], charge_range=[], zoom=None):
        # produce a contour plot
        # Either interpolated or polygon

        # print('plot inputs1:',plot_select,plot_type,axis_type,ring_range,charge_range,zoom)

        # Set inputs not needed for specific plots to default values
        # Argument ELIMINATION doesn't propagate ... while argument object MANIPULATION would propagate back to calling routine
        # This is needed since I am not updating the GUI when updating ring or charge ranges ... so these variables keep
        # values from previous plot calls that may not be relevant
        # I could just reload the GUI to avoid this but I am not sure the overhead is worth it
        if not self.need_ring_number(plot_type):
            ring_range = []

        if not self.need_ionization_state(plot_select):
            charge_range = []

        # print('plot inputs2:',plot_select,plot_type,axis_type,ring_range,charge_range,zoom)

        if plot_type == 'contour-polygon' or plot_type == 'contour-interpolate':
            self.plot_contour(plot_select, plot_type, charge_range, zoom)

        elif plot_type == 'along ring':
            if len(ring_range) == 0:
                print("ERROR: Rings not specified correctly for along ring plots")
            else:
                self.plot_along_ring(plot_select, axis_type, ring_range, charge_range, zoom)

        elif plot_type == 'surface':
            print("Plot not yet implemented")
            return

        elif plot_type == 'diagnostic':
            print("Plot not yet implemented")
            return

        else:
            print("ERROR: Specified plot type: `%s` not found." % plot_type)

    def plot_contour(self, plot_select, plot_type, charge_range=[], zoom=None):
        # This produces polygon contour plots matching the
        # computationals mesh of the data included in the plot list

        # If nplots is 1 then the 'data' reference contains a self[] reference otherwise it contains a list of all_plot dictionary entries
        # set up subplots

        #
        # If charge states are specified for an impurity quantity two things happen
        # 1) The figures go into a figure notebook - one/charge state
        # 2) The get_data_2d routine in the specific contour plotting routines will require the charge state to be specified.
        #

        if len(charge_range) == 0:
            # if no charge states then just plot the figure
            # Note: OMFITx.py Figure is intended to put a figure inside an OMFIT GUI element
            # fig = Figure(returnFigure=True)
            fig = figure()
            self.plot_contour_fig(fig, plot_select, plot_type, charge=None, zoom=zoom)
        else:
            # use a figure notebook
            # For lists of charge states
            from omfit_plot import FigureNotebook

            fig_notebook = FigureNotebook(nfig=0, name='Contour plots by charge state')

            # loop through charge states
            for charge in charge_range:
                fig = fig_notebook.add_figure(label='IZ={}'.format(charge))
                self.plot_contour_fig(fig, plot_select, plot_type, charge, zoom=zoom)
                fig.canvas.draw()

    def plot_contour_fig(self, fig, plot_select, plot_type, charge=None, zoom=None):
        # given the figure - add the appropriate contour plots

        plot_list = OMFIToedgeNC.all_plots[plot_select]['data']
        nplts = len(plot_list)

        # get figure layout for number of plots on the page
        nrows, ncols = self.calculate_layout(nplts)

        for cnt in range(len(plot_list)):
            ax = fig.add_subplot(nrows, ncols, cnt + 1)
            dataname, targname, label, scalef = self.get_names(plot_select, plot_list[cnt])

            if plot_type == 'contour-polygon':
                # print("plot_contour_polygon_call:",scalef)
                self.plot_contour_polygon(fig, ax, dataname, charge, zoom, scalef=scalef)
                ax.set_title(label)

            elif plot_type == 'contour-interpolate':
                self.plot_contour_interpolate(fig, ax, dataname, targname, charge, zoom, scalef=scalef)
                ax.set_title(label)
            # set the figure title
            fig.suptitle(plot_select)

    def plot_contour_polygon(self, fig, ax, dataname, charge=None, zoom=None, scalef=1.0):
        # polygon plots never use the target data
        # produce polygon contour plot
        if charge is None:
            data = self.get_data_2d(dataname, scalef=scalef)
        else:
            data = self.get_data_2d(dataname, targname=None, charge=charge, scalef=scalef)

        coll = matplotlib.collections.PolyCollection(self.grid, array=data, cmap=pyplot.cm.jet, edgecolors='none')
        ax.add_collection(coll)
        ax.plot(self.rves, self.zves, linewidth=1)
        # ax.plot(wall,linewidth=1)
        ax.autoscale_view()
        ax.set_aspect("equal")
        fig.colorbar(coll, ax=ax)

    def plot_contour_interpolate(self, fig, ax, dataname, targname=None, charge=None, zoom=None, scalef=1.0):
        # produce interpolated contour plot
        # not all 2D data has target data
        if charge is None:
            data = self.get_data_2d(dataname, targname, scalef=scalef)
        else:
            data = self.get_data_2d(dataname, targname, charge, scalef=scalef)

        if targname == None:
            r = self.r_cells
            z = self.z_cells
        else:
            r = self.r_points
            z = self.z_points

        triang = matplotlib.tri.Triangulation(r, z)

        # Mask off unwanted triangles by finding ones with center points outside of wall
        # This could be changed to ones with any corner outside wall but then you might get
        # missing triangle gaps inside the wall
        rmid = r[triang.triangles].mean(axis=1)
        zmid = z[triang.triangles].mean(axis=1)
        points = [[rmid[i], zmid[i]] for i in range(len(rmid))]

        res = self.wall_path.contains_points(points)
        mask = np.where(res == False, 1, 0)
        # print("mask",mask)

        # triangles to be plotted should be set now
        triang.set_mask(mask)

        # ax.set_cmap("jet")
        ax.set_aspect('equal')
        ax.plot(self.rves, self.zves, linewidth=1)
        # contour levels are set to 20 but this could be pulled into a settable parameter
        cs = ax.tricontourf(triang, data, 20, cmap=pyplot.cm.jet)
        fig.colorbar(cs, ax=ax)

    def calculate_layout(self, nplts):
        #
        # Calculate a layout that will hold all the figures
        #
        # Handle special cases - then use generic algorithm
        #
        if nplts == 3:
            nrows = 1
            ncols = 3
        elif nplts == 7:
            nrows = 2
            ncols = 4
        else:
            nrows = np.ceil(np.sqrt(nplts))
            ncols = np.ceil(np.sqrt(nplts))
            # if number of plots will fit in a smaller grid remove a row
            if ncols * (nrows - 1) >= nplts:
                nrows = nrows - 1

        # return the layout
        return nrows, ncols

    def get_names(self, plot_select, plot_name):
        # This returns the names from the NC dictionary to be plotted based on the OMFIToedgeNC.all_plots summary
        # It also returns the label associated with the individual plot element
        dataname = None
        targname = None
        label = None
        scale_desc = None

        # each of the dictionary entries contains a list - this list is either a single reference to
        # the NC datafile or a list of reference to other plots. Plot_name should always be a singular plot if it is present

        if plot_select in OMFIToedgeNC.all_plots:
            if plot_name in self:
                # Need to return plot_name as dataname and get targname from the plot select reference
                dataname = plot_name
                if 'targ' in OMFIToedgeNC.all_plots[plot_select]:
                    if OMFIToedgeNC.all_plots[plot_select]['targ'] is not None:
                        targname = OMFIToedgeNC.all_plots[plot_select]['targ'][0]
                # get label
                if 'label' in OMFIToedgeNC.all_plots[plot_select]:
                    if OMFIToedgeNC.all_plots[plot_select]['label'] is not None:
                        label = OMFIToedgeNC.all_plots[plot_select]['label']
                        if 'units' in OMFIToedgeNC.all_plots[plot_select]:
                            if OMFIToedgeNC.all_plots[plot_select]['units'] is not None:
                                label = label + ' ' + OMFIToedgeNC.all_plots[plot_select]['units']
                # get scaling factor if any
                if 'scale' in OMFIToedgeNC.all_plots[plot_select]:
                    if OMFIToedgeNC.all_plots[plot_select]['scale'] is not None:
                        scale_desc = OMFIToedgeNC.all_plots[plot_select]['scale']

            elif plot_name in OMFIToedgeNC.all_plots:
                # Need to load both dataname and targname from the plot list
                if 'data' in OMFIToedgeNC.all_plots[plot_name]:
                    if OMFIToedgeNC.all_plots[plot_name]['data'] is not None:
                        dataname = OMFIToedgeNC.all_plots[plot_name]['data'][0]

                if 'targ' in OMFIToedgeNC.all_plots[plot_name]:
                    if OMFIToedgeNC.all_plots[plot_name]['targ'] is not None:
                        targname = OMFIToedgeNC.all_plots[plot_name]['targ'][0]
                # get data label
                if 'label' in OMFIToedgeNC.all_plots[plot_name]:
                    if OMFIToedgeNC.all_plots[plot_name]['label'] is not None:
                        label = OMFIToedgeNC.all_plots[plot_name]['label']
                        if 'units' in OMFIToedgeNC.all_plots[plot_name]:
                            if OMFIToedgeNC.all_plots[plot_name]['units'] is not None:
                                label = label + ' ' + OMFIToedgeNC.all_plots[plot_name]['units']
                # get scaling factor if any
                if 'scale' in OMFIToedgeNC.all_plots[plot_name]:
                    if OMFIToedgeNC.all_plots[plot_name]['scale'] is not None:
                        scale_desc = OMFIToedgeNC.all_plots[plot_name]['scale']

        #
        #            Comment out - this code is used both to get plots AND to check whether they exist in the nc file
        #
        #            else:
        #                print("ERROR: plot_name must be either a reference to self or a plot in OMFIToedgeNC.all_plots",plot_select,plot_name)

        # set numerical scaling factor for the volumetric data
        # Scaling factor is set 1.0 by default ... only changed if some other quantity is specified
        if scale_desc == 'ABSFAC':
            scalef = self.absfac
        elif scale_desc == '1/QTIM':
            scalef = 1.0 / self.qtim
        else:
            scalef = 1.0

        return dataname, targname, label, scalef

    def get_data_along_ring(self, ir, dataname=None, targname=None, charge=None, scalef=1.0):
        # loads a set of 2D data
        # get along ring data
        # print("get_data_along_ring:",ir,dataname,targname,charge)

        if dataname is not None:
            if charge is None:
                data = self[dataname]['data'][ir][: self.nks[ir]] * scalef
            else:
                # Note: the results array indexing goes from -1:nizs for charge states.
                # -1 maps to 0 in the python array ... charge 0 is in index 1 - so add one to charge
                # to get correct indexing
                data = self[dataname]['data'][charge + 1][ir][: self.nks[ir]] * scalef

        # print("data:",dataname,data)

        # Insert the target data at the beginning and end when it is available
        # Targt data is only available on sol and pfz rings
        if targname is not None and ir >= self.irsep - 1:
            # Target data does not need the scaling
            # get target indices in the target data arrays
            id0 = self.idds[1, ir]
            id1 = self.idds[0, ir]

            # print('id0,id1:',id0,id1,self.idds[1][ir],self.idds[0][ir])
            # print('idds:',self.idds)

            # print('targ:',targname,self[targname]['data'])
            # print('targ2 id:',targname,self.idds[0][:])
            # print('targ1 id:',targname,self.idds[1][:])
            # print('targ data1:',targname,id0,self[targname]['data'][id0])

            data = np.insert(data, 0, [self[targname]['data'][id0]])

            # print('targ data2:',targname,id1,self[targname]['data'][id1])

            data = np.append(data, [self[targname]['data'][id1]])
            # print('data2:',data)

        # return the data
        return data

    def get_axis(self, ir, axis_type='S', targname=None):
        # This loads the particular axis for the along ring plot (either S or P)
        # It also adds the target elements if targname is not None
        if axis_type == 'P':
            axis = self.kps[ir][: self.nks[ir]]
            label = 'P (m)'
            if targname is not None and ir > self.irsep - 1:
                axis = np.insert(axis, 0, 0.0)
                axis = np.append(axis, self.pmax[ir])
        else:  #  if axis_type == 'S':  # if someone forgets to set axis_type treat it as 'S'
            axis = self.kss[ir][: self.nks[ir]]
            label = 'S (m)'
            if targname is not None and ir >= self.irsep - 1:
                axis = np.insert(axis, 0, 0.0)
                axis = np.append(axis, self.smax[ir])

        return axis, label

    def plot_ring(self, ir, fig, plot_select, axis_type='S', charge_range=[]):
        # plot selected data for specified ring on figure specified

        if ir < 1 or ir > self.nrs:
            print("ERROR: Ring specified is out of range available: %d" % ir)
            return
        else:
            # ring references are in 1 .. nrs from fortran while python arrays index from zero
            ir_ref = ir - 1

        plot_list = OMFIToedgeNC.all_plots[plot_select]['data']
        nplts = len(plot_list)

        # If nplots is 1 then the 'data' reference contains a self[] reference otherwise it contains a list of all_plot dictionary entries

        # set up subplots

        nrows, ncols = self.calculate_layout(nplts)

        for cnt in range(len(plot_list)):
            ax = fig.add_subplot(nrows, ncols, cnt + 1)
            if len(charge_range) == 0:
                dataname, targname, label, scalef = self.get_names(plot_select, plot_list[cnt])
                axis, x_label = self.get_axis(ir_ref, axis_type, targname)
                data = self.get_data_along_ring(ir_ref, dataname, targname, scalef=scalef)

                # add a marker for the end points if targname is not None
                if targname is not None:
                    markers_on = [0, len(axis) - 1]
                else:
                    markers_on = []
                # plot the data
                ax.plot(axis, data, linestyle='-', marker='o', markevery=markers_on, linewidth=1)
                ax.set_xlabel(x_label)
                ax.set_ylabel(label)
                ax.set_xlim(0.0, axis[-1])
            else:
                for charge in charge_range:
                    dataname, targname, label, scalef = self.get_names(plot_select, plot_list[cnt])
                    axis, x_label = self.get_axis(ir_ref, axis_type, targname)
                    data = self.get_data_along_ring(ir_ref, dataname, targname, charge, scalef=scalef)
                    # add a marker for the end points if targname is not None
                    if targname is not None:
                        markers_on = [0, len(axis) - 1]
                    else:
                        markers_on = []
                    # plot the data
                    ax.plot(axis, data, label='IZ={}'.format(charge), linestyle='-', marker='o', markevery=markers_on, linewidth=1)
                    ax.set_xlabel(x_label)
                    ax.set_ylabel(label)
                    ax.set_xlim(0.0, axis[-1])
                ax.legend()

    def plot_along_ring(self, plot_select, axis_type, ring_range, charge_range=[], zoom=None):
        # Routine plots the quantities along ring ... number of plots/page is controlled by number
        # of items in the plot list.

        # Single figures
        # Single ring/single quantity
        # single ring/multiple quantities
        #
        # Figure notebook spanning ring range
        # multi rings/single quantity
        # multi rings/multiple quantities
        #
        # If more than one ring is specified then a figure notebook is used
        #
        # If more than one charge - try to put them on the same axes
        #
        # plot

        nrings = len(ring_range)
        if nrings <= 0:
            # error condition
            print("ERROR: No rings specified for the plot")
            return

        minring = 1
        maxring = self.nrs

        # print('rings:',nrings,ring_range)

        res = [ring_range[i] < minring or ring_range[i] > maxring for i in range(len(ring_range))]

        # print('res:',res)

        out_of_bounds = np.where(res == True)
        # print('out_of_bounds:',len(out_of_bounds),)
        # print('test:',ring_range[out_of_bounds])

        # Check to see if at least some rings are in range
        if len(out_of_bounds) >= nrings:
            print("ERROR: No valid rings numbers specified: min=%d max=%d range=%s" % (minring, maxring, str(ring_range)))
            return

        plot_list = OMFIToedgeNC.all_plots[plot_select]['data']
        nplts = len(plot_list)

        # if nplts is 1 then the list contains a self reference ... if greater than 1 then it contains a list of plot references
        # Do everything in a figure notebook since there can be 1+ figures generated

        # print('rings:',nrings,ring_range,plot_select,axis_type)

        # use figure notebook since it supports an unknown number of plots
        from omfit_plot import FigureNotebook

        fig_notebook = FigureNotebook(nfig=0, name='Along Ring Plots')

        for ir in ring_range:
            fig = fig_notebook.add_figure(label='IR={}'.format(ir))
            self.plot_ring(ir, fig, plot_select, axis_type, charge_range)
            fig.canvas.draw()



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import *
from omfit_classes.omfit_weblink import OMFITwebLink
import numpy as np

__all__ = ['OMFITbibtex', 'searchdoi']

try:
    import bibtexparser
except Exception as _excp:
    bibtexparser = None
    printe('WARNING! Compromised omfit_bibtex support due to failure to import bibtexparser!')


class OMFITbibtex(OMFITascii, SortedDict):
    r"""
    Class used to parse bibtex files
    The class should be saved as a dictionary or dictionaries (one dictionary for each bibtex entry)
    Each bibtex entry must have defined the keys: `ENTRYTYPE` and `ID`

    :param filename: filename of the .bib file to parse

    :param \**kw: keyword dictionary passed to OMFITascii class

    To generate list of own publications:
    1. Export all of your citations from https://scholar.google.com to a `citation.bib` bibtex file
    2. OMFIT['bib']=OMFITbibtex('.bib')                   # load citations as OMFITbibtex
    3. OMFIT['bib'].doi(deleteNoDOI=True)                 # remove entries which do not have a DOI (ie.conferences)
    4. OMFIT['bib'].sanitize()                            # fix entries where needed
    4. OMFIT['bib'].update_ID(as_author='Meneghini')      # Sort entries and distinguish between first author or contributed
    5. print('\n\n'.join(OMFIT['bib'].write_format()))    # write to whatever format desired

    """

    def __init__(self, filename, **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self, caseInsensitive=True)
        if filename is None:
            from datetime import date

            self[0] = {'ID': os.environ['USER'] + str(date.today().year), 'ENTRYTYPE': 'article'}
        else:
            self.dynaLoad = True

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            bibtex_str = f.read()
        bib_database = bibtexparser.loads(bibtex_str)
        for k, item in enumerate(bib_database.get_entry_list()):
            label0 = label = item['ID']
            k = 0
            while label in self:
                k += 1
                label = label0 + '_%d' % k
            self[label] = item

    def write_format(self, form='\\item[]{{{author}; {title}; {journal} {volume} {year}: \\href{{http://dx.doi.org/{doi}}}{{doi:{doi}}}}}'):
        """
        returns list with entries formatted according to `form` string

        :param form: format to use to

        :return: list of strings
        """
        txt = []
        for item in list(self.keys()):
            tmp = self[item].copy()
            tmp['author'] = re.sub('others', 'et al.', ', '.join([re.sub(', *', ' ', x) for x in tmp['author'].split(' and ')]))
            tmp['volume'] = tmp.get('volume', '1')
            txt.append(form.format(**tmp))
        return txt

    @dynaSave
    def save(self):
        tmp = bibtexparser.bibdatabase.BibDatabase()
        tmp.entries = list(self.values())
        writer = bibtexparser.bwriter.BibTexWriter()
        bibtex_str = encode_ascii_ignore(bibtexparser.dumps(tmp))
        if self.filename is not None:
            with open(self.filename, 'w') as f:
                f.write(bibtex_str)
        else:
            return bibtex_str

    def doi(self, deleteNoDOI=False):
        """
        method for adding DOI information to bibtex

        :param deleteNoDOI: delete entries without DOI
        """
        for item in list(self.keys()):
            # only articles
            if (
                'doi' in self[item]
                or np.any([k not in self[item] for k in ['ENTRYTYPE', 'author', 'title', 'journal']])
                or self[item]['journal'].lower() in [x.lower() for x in _noDOIjournals]
                or self[item]['ENTRYTYPE'].lower() != 'article'
            ):
                printi('%s: %s' % (encode_ascii_ignore(item), self[item].get('doi', '')))
                continue

            # search for this paper
            tmp = searchdoi(self[item]['title'], self[item]['author'])
            self[item]['__doi__'] = tmp
            self[item]['doi'] = tmp[0]['DOI']
            self[item]['doi_score'] = tmp[0]['score']

            if self[item]['title'].lower() != tmp[0]['title'][0].lower():
                # if the title does not match exactly this is a bad sign
                printw('%s: %s' % (encode_ascii_ignore(item), self[item]['doi']))
            else:
                if np.all([k in tmp[0] for k in ['container-title', 'issue', 'page', 'publisher']]):
                    # trust the info from crossref.org
                    self[item]['journal'] = tmp[0]['container-title'][0]
                    self[item]['number'] = tmp[0]['issue']
                    self[item]['pages'] = tmp[0]['page']
                    self[item]['publisher'] = tmp[0]['publisher']
                    self[item]['author'] = ' and '.join(
                        [', '.join([_f for _f in [x.get('family', 'others'), x.get('given', '')] if _f]) for x in tmp[0]['author']]
                    )
                printi('%s: %s' % (encode_ascii_ignore(item), self[item]['doi']))

        # create URL entry based on DOI
        for item in list(self.keys()):
            if 'doi' in self[item]:
                self[item]['url'] = OMFITwebLink('doi.org/' + self[item]['doi'])
            elif deleteNoDOI:
                del self[item]

        # finar clearnup
        for item in list(self.keys()):
            if '__doi__' in self[item]:
                del self[item]['__doi__']
            if 'doi_score' in self[item]:
                del self[item]['doi_score']

    def sanitize(self):
        """
        Sanitizes the database entries:
        1. Fix all-caps author names
        2. Fix unicodes
        """
        for item in self:
            # Fix all-caps author names
            authors = self[item]['author'].split()
            for k, a in enumerate(authors):
                if '.' in a or len(a) > 2:
                    authors[k] = a.title()
            self[item]['author'] = ' '.join(authors)
            # Fix unicodes
            for sub in list(self[item].keys()):
                self[item][sub] = encode_ascii_ignore(self[item][sub])

    def filter(self, conditions):
        """
        filter database given a set of conditions

        :param conditions: list of strings (eg. ['int(year)>2012']

        :return: filtered OMFITbibtex object
        """
        conditions = tolist(conditions)
        self = copy.deepcopy(self)
        for condition in conditions:
            for item in list(self.keys()):
                try:
                    if not eval(condition, {}, self[item]):
                        del self[item]
                except Exception:
                    del self[item]
        return self

    def update_ID(self, fmt=['year_1stAuthor_jrnl', 'lower1stAuthor_year'][1], separator=':', as_author=False):
        """
        set bibtex ID

        :param fmt: string with format 'year_1stAuthor_jrnl'

        :param separator: string with separator for fmt

        :param as_author: only keep entries that have `as_author` as author
        """
        for item in list(self.keys()):
            tmp = self[item]
            del self[item]
            try:
                tmp['1stAuthor'] = tmp['author'].split()[0].strip(',')
                tmp['lower1stAuthor'] = tmp['1stAuthor'].lower()
            except Exception:
                pass
            try:
                tmp['jrnl'] = ''.join([x[0] for x in tmp['journal'].split()])
                tmp['lowerjrnl'] = tmp['jrnl'].lower()
            except Exception:
                pass
            label0 = label = separator.join([str(tmp[k]) for k in fmt.split('_')])
            if as_author:
                if as_author.lower() not in tmp['author'].lower():
                    continue
                if tmp['author'].split()[0].strip(',').lower() == as_author.lower():
                    label = label0 = '_' + label
            k = 0
            while label in self:
                k += 1
                label = label0 + '%s' % chr(k + ord('a'))
            for item in ['1stAuthor', 'jrnl', 'lower1stAuthor', 'lowerjrnl']:
                if item in tmp:
                    del tmp[item]
            self[label] = tmp
            self[label]['ID'] = label
        self.sort()
        return self


def searchdoi(title, author):
    """
    This function returns a list of dictionaries containing the best matching papers
    for the title and authors according to the crossref.org website

    :param title: string with the title

    :param author: string with the authors

    :return: list of dictionaries with info about the papers found
    """
    data = {"query.author": author.encode('utf-8'), "query.title": title.encode('utf-8')}
    url = "http://api.crossref.org/works?"

    import requests
    import json

    response = requests.get(url, data, verify=False)

    try:
        response.raise_for_status()
        return json.loads(response.text)['message']['items']
    except Exception:
        raise


_noDOIjournals = ['Bulletin of the American Physical Society', 'Bull. Am. Phys. Soc.']

############################################
if __name__ == '__main__':
    test_classes_main_header()
    import requests

    bibtex_text = '''
@article{meneghini2013integrated,
  title={Integrated modeling of tokamak experiments with OMFIT},
  author={Meneghini, Orso and Lao, Lang},
  journal={Plasma and Fusion Research},
  volume={8},
  pages={2403009--2403009},
  year={2013},
  publisher={The Japan Society of Plasma Science and Nuclear Fusion Research}
}

@article{meneghini2015integrated,
  title={Integrated modeling applications for tokamak experiments with OMFIT},
  author={Meneghini, O and Smith, SP and Lao, LL and Izacard, O and Ren, Q and Park, JM and Candy, J and Wang, Z and Luna, CJ and Izzo, VA and others},
  journal={Nuclear Fusion},
  volume={55},
  number={8},
  pages={083008},
  year={2015},
  publisher={IOP Publishing}
}
'''

    bib = OMFITbibtex('OMFIT_bibtex', fromString=bibtex_text)
    # filter
    filtered_bib = bib.filter(['int(year)>2014'])
    # sanitize
    filtered_bib.sanitize()
    # get the DOI (can take some time, and is thus best done after filtering)
    try:
        bib.doi()
    except requests.exceptions.HTTPError:
        pass
    pprint(filtered_bib)
